@@ -1,9 +1,6 @@
 module Hedgehog.Gen (
-  -- * Combinators
-    lift
-
   -- ** Shrinking
-  , shrink
+    shrink
   , prune
 
   -- ** Size
@@ -70,6 +67,7 @@ module Hedgehog.Gen (
   -- ** Conditional
   , discard
   , filter
+  , mapMaybe
   , just
 
   -- ** Collections
