@@ -1,4 +1,6 @@
 {-# OPTIONS_HADDOCK not-home #-}
+{-# LANGUAGE ApplicativeDo #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
 {-# LANGUAGE DeriveFoldable #-}
 {-# LANGUAGE DeriveFunctor #-}
@@ -16,6 +18,9 @@
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE UndecidableInstances #-} -- MonadBase
+#if __GLASGOW_HASKELL__ >= 806
+{-# LANGUAGE DerivingVia #-}
+#endif
 module Hedgehog.Internal.Gen (
   -- * Transformer
     Gen
@@ -23,7 +28,7 @@ module Hedgehog.Internal.Gen (
   , MonadGen(..)
 
   -- * Combinators
-  , lift
+  , generalize
 
   -- ** Shrinking
   , shrink
@@ -94,7 +99,11 @@ module Hedgehog.Internal.Gen (
   , discard
   , ensure
   , filter
+  , mapMaybe
+  , filterT
+  , mapMaybeT
   , just
+  , justT
 
   -- ** Collections
   , maybe
@@ -116,6 +125,7 @@ module Hedgehog.Internal.Gen (
   -- ** Combinations & Permutations
   , subsequence
   , shuffle
+  , shuffleSeq
 
   -- * Sampling Generators
   , sample
@@ -123,16 +133,24 @@ module Hedgehog.Internal.Gen (
   , printTree
   , printWith
   , printTreeWith
+  , renderTree
 
   -- * Internal
   -- $internal
 
   -- ** Transfomer
   , runGenT
+  , evalGen
+  , evalGenT
   , mapGenT
   , generate
-  , liftTree
+  , toTree
+  , toTreeMaybeT
+  , fromTree
+  , fromTreeT
+  , fromTreeMaybeT
   , runDiscardEffect
+  , runDiscardEffectT
 
   -- ** Size
   , golden
@@ -148,40 +166,42 @@ module Hedgehog.Internal.Gen (
   , Vec(..)
   , Nat(..)
   , subtermMVec
-
-  -- ** Sampling
-  , renderNodes
   ) where
 
 import           Control.Applicative (Alternative(..),liftA2)
-import           Control.Monad (MonadPlus(..), filterM, replicateM, ap, join)
+import           Control.Monad (MonadPlus(..), filterM, guard, replicateM, join)
 import           Control.Monad.Base (MonadBase(..))
+import           Control.Monad.Trans.Control (MonadBaseControl(..))
 import           Control.Monad.Catch (MonadThrow(..), MonadCatch(..))
 import           Control.Monad.Error.Class (MonadError(..))
+import           Control.Monad.Fail (MonadFail (..))
+import qualified Control.Monad.Fail as Fail
 import           Control.Monad.IO.Class (MonadIO(..))
-import           Control.Monad.Morph (MFunctor(..), MMonad(..), generalize)
+import           Control.Monad.Morph (MFunctor(..), MMonad(..))
+import qualified Control.Monad.Morph as Morph
 import           Control.Monad.Primitive (PrimMonad(..))
 import           Control.Monad.Reader.Class (MonadReader(..))
 import           Control.Monad.State.Class (MonadState(..))
-import           Control.Monad.Trans.Class (MonadTrans)
-import qualified Control.Monad.Trans.Class as Trans
-import           Control.Monad.Trans.Except (ExceptT(..), mapExceptT)
-import           Control.Monad.Trans.Identity (IdentityT(..), mapIdentityT)
-import           Control.Monad.Trans.Maybe (MaybeT(..), mapMaybeT)
-import qualified Control.Monad.Trans.RWS.Lazy as Lazy
-import qualified Control.Monad.Trans.RWS.Strict as Strict
-import           Control.Monad.Trans.Reader (ReaderT(..), mapReaderT)
+import           Control.Monad.Trans.Class (MonadTrans(..))
+import           Control.Monad.Trans.Except (ExceptT(..))
+import           Control.Monad.Trans.Identity (IdentityT(..))
+import           Control.Monad.Trans.Maybe (MaybeT(..))
+import           Control.Monad.Trans.Reader (ReaderT(..))
 import           Control.Monad.Trans.Resource (MonadResource(..))
 import qualified Control.Monad.Trans.State.Lazy as Lazy
 import qualified Control.Monad.Trans.State.Strict as Strict
 import qualified Control.Monad.Trans.Writer.Lazy as Lazy
 import qualified Control.Monad.Trans.Writer.Strict as Strict
 import           Control.Monad.Writer.Class (MonadWriter(..))
+import           Control.Monad.Zip (MonadZip(..))
 
-import           Data.Bifunctor (first, second)
+import           Data.Bifunctor (first)
 import           Data.ByteString (ByteString)
 import qualified Data.ByteString as ByteString
 import qualified Data.Char as Char
+#if __GLASGOW_HASKELL__ < 806
+import           Data.Coerce (coerce)
+#endif
 import           Data.Foldable (for_, toList)
 import           Data.Functor.Identity (Identity(..))
 import           Data.Int (Int8, Int16, Int32, Int64)
@@ -200,11 +220,11 @@ import qualified Data.Text as Text
 import qualified Data.Text.Encoding as Text
 import           Data.Word (Word8, Word16, Word32, Word64)
 
-import           Hedgehog.Internal.Distributive (Distributive(..))
+import           Hedgehog.Internal.Distributive (MonadTransDistributive(..))
 import           Hedgehog.Internal.Seed (Seed)
 import qualified Hedgehog.Internal.Seed as Seed
 import qualified Hedgehog.Internal.Shrink as Shrink
-import           Hedgehog.Internal.Tree (Tree(..), Node(..))
+import           Hedgehog.Internal.Tree (Tree, TreeT(..), NodeT(..))
 import qualified Hedgehog.Internal.Tree as Tree
 import           Hedgehog.Range (Size, Range)
 import qualified Hedgehog.Range as Range
@@ -224,18 +244,34 @@ type Gen =
 --
 newtype GenT m a =
   GenT {
-      unGen :: Size -> Seed -> Tree (MaybeT m) a
+      unGenT :: Size -> Seed -> TreeT (MaybeT m) a
     }
 
 -- | Runs a generator, producing its shrink tree.
 --
-runGenT :: Size -> Seed -> GenT m a -> Tree (MaybeT m) a
+runGenT :: Size -> Seed -> GenT m a -> TreeT (MaybeT m) a
 runGenT size seed (GenT m) =
   m size seed
 
+-- | Run a generator, producing its shrink tree.
+--
+--   'Nothing' means discarded, 'Just' means we have a value.
+--
+evalGen :: Size -> Seed -> Gen a -> Maybe (Tree a)
+evalGen size seed =
+  Tree.mapMaybe id .
+  evalGenT size seed
+
+-- | Runs a generator, producing its shrink tree.
+--
+evalGenT :: Monad m => Size -> Seed -> GenT m a -> TreeT m (Maybe a)
+evalGenT size seed =
+  runDiscardEffectT .
+  runGenT size seed
+
 -- | Map over a generator's shrink tree.
 --
-mapGenT :: (Tree (MaybeT m) a -> Tree (MaybeT n) b) -> GenT m a -> GenT n b
+mapGenT :: (TreeT (MaybeT m) a -> TreeT (MaybeT n) b) -> GenT m a -> GenT n b
 mapGenT f gen =
   GenT $ \size seed ->
     f (runGenT size seed gen)
@@ -243,308 +279,236 @@ mapGenT f gen =
 -- | Lift a predefined shrink tree in to a generator, ignoring the seed and the
 --   size.
 --
-liftTree :: Tree (MaybeT m) a -> GenT m a
-liftTree x =
-  GenT (\_ _ -> x)
+fromTree :: MonadGen m => Tree a -> m a
+fromTree =
+  fromTreeT .
+  hoist (Morph.generalize)
+
+-- | Lift a predefined shrink tree in to a generator, ignoring the seed and the
+--   size.
+--
+fromTreeT :: MonadGen m => TreeT (GenBase m) a -> m a
+fromTreeT x =
+  fromTreeMaybeT $
+    hoist (MaybeT . fmap Just) x
+
+-- | Lift a predefined shrink tree in to a generator, ignoring the seed and the
+--   size.
+--
+fromTreeMaybeT :: MonadGen m => TreeT (MaybeT (GenBase m)) a -> m a
+fromTreeMaybeT x =
+  fromGenT . GenT $ \_ _ ->
+    x
+
+-- | Observe a generator's shrink tree.
+--
+toTree :: forall m a. (MonadGen m, GenBase m ~ Identity) => m a -> m (Tree a)
+toTree =
+  withGenT $ mapGenT (Maybe.maybe empty pure . runDiscardEffect)
+
+-- | Lift a predefined shrink tree in to a generator, ignoring the seed and the
+--   size.
+--
+toTreeMaybeT :: MonadGen m => m a -> m (TreeT (MaybeT (GenBase m)) a)
+toTreeMaybeT =
+  withGenT $ mapGenT pure
+
+-- | Lazily run the discard effects through the tree and reify it a
+--   @Maybe (Tree a)@.
+--
+--   'Nothing' means discarded, 'Just' means we have a value.
+--
+--   Discards in the child nodes of the tree are simply removed.
+--
+runDiscardEffect :: TreeT (MaybeT Identity) a -> Maybe (Tree a)
+runDiscardEffect =
+  Tree.mapMaybe id .
+  runDiscardEffectT
 
 -- | Run the discard effects through the tree and reify them as 'Maybe' values
---   at the nodes. 'Nothing' means discarded, 'Just' means we have a value.
+--   at the nodes.
 --
-runDiscardEffect :: Monad m => Tree (MaybeT m) a -> Tree m (Maybe a)
-runDiscardEffect =
-  runMaybeT . distribute
+--   'Nothing' means discarded, 'Just' means we have a value.
+--
+runDiscardEffectT :: Monad m => TreeT (MaybeT m) a -> TreeT m (Maybe a)
+runDiscardEffectT =
+  runMaybeT .
+  distributeT
+
+-- | Lift a @Gen / GenT Identity@ in to a @Monad m => GenT m@
+--
+generalize :: Monad m => Gen a -> GenT m a
+generalize =
+  hoist Morph.generalize
 
 ------------------------------------------------------------------------
 -- MonadGen
 
 -- | Class of monads which can generate input data for tests.
 --
---   /The functions on this class can, and should, be used without their @Gen@/
---   /suffix by importing "Hedgehog.Gen" qualified./
+class (Monad m, Monad (GenBase m)) => MonadGen m where
+  type GenBase m :: (* -> *)
+
+  -- | Extract a 'GenT' from a  'MonadGen'.
+  --
+  toGenT :: m a -> GenT (GenBase m) a
+
+  -- | Lift a 'GenT' in to a 'MonadGen'.
+  --
+  fromGenT :: GenT (GenBase m) a -> m a
+
+-- | Transform a 'MonadGen' as a 'GenT'.
 --
-class Monad m => MonadGen m where
-  -- | See @Gen.@'Hedgehog.Gen.lift'
-  --
-  liftGen :: Gen a -> m a
-
-  -- | See @Gen.@'Hedgehog.Gen.shrink'
-  --
-  shrinkGen :: (a -> [a]) -> m a -> m a
-
-  -- | See @Gen.@'Hedgehog.Gen.prune'
-  --
-  pruneGen :: m a -> m a
-
-  -- | See @Gen.@'Hedgehog.Gen.scale'
-  --
-  scaleGen :: (Size -> Size) -> m a -> m a
-
-  -- | See @Gen.@'Hedgehog.Gen.freeze'
-  --
-  freezeGen :: m a -> m (a, m a)
+withGenT :: (MonadGen m, MonadGen n) => (GenT (GenBase m) a -> GenT (GenBase n) b) -> m a -> n b
+withGenT f =
+  fromGenT . f . toGenT
 
 instance Monad m => MonadGen (GenT m) where
-  liftGen gen =
-    hoist generalize gen
+  -- | The type of the transformer stack's base 'Monad'.
+  --
+  type GenBase (GenT m) =
+    m
 
-  shrinkGen =
-    mapGenT . Tree.expand
+  -- | Convert a 'MonadGen' to a 'GenT'.
+  --
+  toGenT =
+    id
 
-  pruneGen =
-    mapGenT Tree.prune
-
-  scaleGen f gen =
-    GenT $ \size0 seed ->
-      let
-        size =
-          f size0
-      in
-        if size < 0 then
-          error "Hedgehog.Gen.scale: negative size"
-        else
-          runGenT size seed gen
-
-  freezeGen gen =
-    GenT $ \size seed -> do
-      mx <- Trans.lift . Trans.lift . runMaybeT . runTree $ runGenT size seed gen
-      case mx of
-        Nothing ->
-          mzero
-        Just (Node x xs) ->
-          pure (x, liftTree . Tree.fromNode $ Node x xs)
+  -- | Convert a 'GenT' to a 'MonadGen'.
+  --
+  fromGenT =
+    id
 
 instance MonadGen m => MonadGen (IdentityT m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (IdentityT m) =
+    IdentityT (GenBase m)
 
-  shrinkGen f =
-    mapIdentityT (shrink f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen =
-    mapIdentityT $
-      fmap (second Trans.lift) . freeze
-
-shrinkMaybe :: (a -> [a]) -> Maybe a -> [Maybe a]
-shrinkMaybe f = \case
-  Nothing ->
-    pure Nothing
-  Just x ->
-    fmap Just (f x)
-
-shrinkEither :: (a -> [a]) -> Either x a -> [Either x a]
-shrinkEither f = \case
-  Left x ->
-    pure $ Left x
-  Right x ->
-    fmap Right (f x)
-
-shrink2 :: (a -> [a]) -> (a, b) -> [(a, b)]
-shrink2 f (x, y) =
-  fmap (, y) (f x)
-
-shrink3 :: (a -> [a]) -> (a, b, c) -> [(a, b, c)]
-shrink3 f (x, y, z) =
-  fmap (, y, z) (f x)
+  fromGenT =
+    hoist fromGenT . distributeT
 
 instance MonadGen m => MonadGen (MaybeT m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (MaybeT m) =
+    MaybeT (GenBase m)
 
-  shrinkGen f =
-    mapMaybeT $
-      shrink (shrinkMaybe f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen =
-    mapMaybeT $ \m0 -> do
-      (mx, m) <- freeze m0
-      pure $ fmap (, MaybeT m) mx
+  fromGenT =
+    hoist fromGenT . distributeT
 
 instance MonadGen m => MonadGen (ExceptT x m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (ExceptT x m) =
+    ExceptT x (GenBase m)
 
-  shrinkGen f =
-    mapExceptT $
-      shrink (shrinkEither f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen =
-    mapExceptT $ \m0 -> do
-      (mx, m) <- freeze m0
-      pure $ fmap (, ExceptT m) mx
+  fromGenT =
+    hoist fromGenT . distributeT
 
 instance MonadGen m => MonadGen (ReaderT r m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (ReaderT r m) =
+    ReaderT r (GenBase m)
 
-  shrinkGen f =
-    mapReaderT (shrink f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
+  fromGenT =
+    hoist fromGenT . distributeT
 
-  scaleGen f =
-    hoist (scale f)
+instance MonadGen m => MonadGen (Lazy.StateT r m) where
+  type GenBase (Lazy.StateT r m) =
+    Lazy.StateT r (GenBase m)
 
-  freezeGen =
-    mapReaderT $
-      fmap (second Trans.lift) . freeze
+  toGenT =
+    distributeT . hoist toGenT
 
-instance MonadGen m => MonadGen (Lazy.StateT s m) where
-  liftGen =
-    Trans.lift . liftGen
+  fromGenT =
+    hoist fromGenT . distributeT
 
-  shrinkGen f =
-    Lazy.mapStateT $
-      shrink (shrink2 f)
+instance MonadGen m => MonadGen (Strict.StateT r m) where
+  type GenBase (Strict.StateT r m) =
+    Strict.StateT r (GenBase m)
 
-  pruneGen =
-    hoist prune
+  toGenT =
+    distributeT . hoist toGenT
 
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Lazy.StateT $ \s0 -> do
-      ((x, s), m) <- freeze (Lazy.runStateT m0 s0)
-      pure ((x, Lazy.StateT (const m)), s)
-
-instance MonadGen m => MonadGen (Strict.StateT s m) where
-  liftGen =
-    Trans.lift . liftGen
-
-  shrinkGen f =
-    Strict.mapStateT $
-      shrink (shrink2 f)
-
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Strict.StateT $ \s0 -> do
-      ((x, s), m) <- freeze (Strict.runStateT m0 s0)
-      pure ((x, Strict.StateT (const m)), s)
+  fromGenT =
+    hoist fromGenT . distributeT
 
 instance (MonadGen m, Monoid w) => MonadGen (Lazy.WriterT w m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (Lazy.WriterT w m) =
+    Lazy.WriterT w (GenBase m)
 
-  shrinkGen f =
-    Lazy.mapWriterT $
-      shrink (shrink2 f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Lazy.WriterT $ do
-      ((x, w), m) <- freeze (Lazy.runWriterT m0)
-      pure ((x, Lazy.WriterT m), w)
+  fromGenT =
+    hoist fromGenT . distributeT
 
 instance (MonadGen m, Monoid w) => MonadGen (Strict.WriterT w m) where
-  liftGen =
-    Trans.lift . liftGen
+  type GenBase (Strict.WriterT w m) =
+    Strict.WriterT w (GenBase m)
 
-  shrinkGen f =
-    Strict.mapWriterT $
-      shrink (shrink2 f)
+  toGenT =
+    distributeT . hoist toGenT
 
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Strict.WriterT $ do
-      ((x, w), m) <- freeze (Strict.runWriterT m0)
-      pure ((x, Strict.WriterT m), w)
-
-instance (MonadGen m, Monoid w) => MonadGen (Lazy.RWST r w s m) where
-  liftGen =
-    Trans.lift . liftGen
-
-  shrinkGen f =
-    Lazy.mapRWST $
-      shrink (shrink3 f)
-
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Lazy.RWST $ \r s0 -> do
-      ((x, s, w), m) <- freeze (Lazy.runRWST m0 r s0)
-      pure ((x, Lazy.RWST (\_ _ -> m)), s, w)
-
-instance (MonadGen m, Monoid w) => MonadGen (Strict.RWST r w s m) where
-  liftGen =
-    Trans.lift . liftGen
-
-  shrinkGen f =
-    Strict.mapRWST $
-      shrink (shrink3 f)
-
-  pruneGen =
-    hoist prune
-
-  scaleGen f =
-    hoist (scale f)
-
-  freezeGen m0 =
-    Strict.RWST $ \r s0 -> do
-      ((x, s, w), m) <- freeze (Strict.runRWST m0 r s0)
-      pure ((x, Strict.RWST (\_ _ -> m)), s, w)
+  fromGenT =
+    hoist fromGenT . distributeT
 
 ------------------------------------------------------------------------
 -- GenT instances
 
 instance (Monad m, Semigroup a) => Semigroup (GenT m a) where
-  (<>) = liftA2 (Semigroup.<>)
+  (<>) =
+    liftA2 (Semigroup.<>)
 
 instance (Monad m, Monoid a) => Monoid (GenT m a) where
-  mappend = liftA2 mappend
-  mempty = return mempty
+  mappend =
+    liftA2 mappend
+
+  mempty =
+    return mempty
 
 instance Functor m => Functor (GenT m) where
   fmap f gen =
     GenT $ \seed size ->
       fmap f (runGenT seed size gen)
 
+--
+-- implementation: parallel shrinking
+--
 instance Monad m => Applicative (GenT m) where
   pure =
-    return
-  (<*>) =
-    ap
+    fromTreeMaybeT . pure
+
+  (<*>) f m =
+    GenT $ \ size seed ->
+      case Seed.split seed of
+        (sf, sm) ->
+          uncurry ($) <$>
+            runGenT size sf f `mzip`
+            runGenT size sm m
+
+--
+-- implementation: satisfies law (ap = <*>)
+--
+--instance Monad m => Applicative (GenT m) where
+--  pure =
+--    fromTreeMaybeT . pure
+--  (<*>) f m =
+--    GenT $ \ size seed ->
+--      case Seed.split seed of
+--        (sf, sm) ->
+--          runGenT size sf f <*>
+--          runGenT size sm m
 
 instance Monad m => Monad (GenT m) where
   return =
-    liftTree . pure
+    pure
 
   (>>=) m k =
     GenT $ \size seed ->
@@ -553,15 +517,26 @@ instance Monad m => Monad (GenT m) where
           runGenT size sk . k =<<
           runGenT size sm m
 
+#if MIN_VERSION_base(4,13,0)
+#else
+  fail =
+    Fail.fail
+#endif
+
+instance Monad m => MonadFail (GenT m) where
+  fail =
+    error
+
 instance Monad m => Alternative (GenT m) where
   empty =
     mzero
+
   (<|>) =
     mplus
 
 instance Monad m => MonadPlus (GenT m) where
   mzero =
-    liftTree mzero
+    fromTreeMaybeT mzero
 
   mplus x y =
     GenT $ \size seed ->
@@ -572,70 +547,97 @@ instance Monad m => MonadPlus (GenT m) where
 
 instance MonadTrans GenT where
   lift =
-    liftTree . Trans.lift . Trans.lift
+    fromTreeMaybeT . lift . lift
 
 instance MFunctor GenT where
   hoist f =
     mapGenT (hoist (hoist f))
 
-embedMaybe ::
+embedMaybeT ::
      MonadTrans t
   => Monad n
   => Monad (t (MaybeT n))
   => (forall a. m a -> t (MaybeT n) a)
   -> MaybeT m b
   -> t (MaybeT n) b
-embedMaybe f m =
-  Trans.lift . MaybeT . pure =<< f (runMaybeT m)
+embedMaybeT f m =
+  lift . MaybeT . pure =<< f (runMaybeT m)
 
-embedTree :: Monad n => (forall a. m a -> Tree (MaybeT n) a) -> Tree (MaybeT m) b -> Tree (MaybeT n) b
-embedTree f tree =
-  embed (embedMaybe f) tree
+embedTreeMaybeT ::
+     Monad n
+  => (forall a. m a -> TreeT (MaybeT n) a)
+  -> TreeT (MaybeT m) b
+  -> TreeT (MaybeT n) b
+embedTreeMaybeT f tree_ =
+  embed (embedMaybeT f) tree_
 
-embedGen :: Monad n => (forall a. m a -> GenT n a) -> GenT m b -> GenT n b
-embedGen f gen =
+embedGenT ::
+     Monad n
+  => (forall a. m a -> GenT n a)
+  -> GenT m b
+  -> GenT n b
+embedGenT f gen =
   GenT $ \size seed ->
     case Seed.split seed of
       (sf, sg) ->
-        (runGenT size sf . f) `embedTree`
+        (runGenT size sf . f) `embedTreeMaybeT`
         (runGenT size sg gen)
 
 instance MMonad GenT where
   embed =
-    embedGen
+    embedGenT
 
-distributeGen :: Transformer t GenT m => GenT (t m) a -> t (GenT m) a
-distributeGen x =
-  join . Trans.lift . GenT $ \size seed ->
-    pure . hoist liftTree . distribute . hoist distribute $ runGenT size seed x
+distributeGenT :: Transformer t GenT m => GenT (t m) a -> t (GenT m) a
+distributeGenT x =
+  join . lift . GenT $ \size seed ->
+    pure . hoist fromTreeMaybeT . distributeT . hoist distributeT $ runGenT size seed x
 
-instance Distributive GenT where
+instance MonadTransDistributive GenT where
   type Transformer t GenT m = (
       Monad (t (GenT m))
     , Transformer t MaybeT m
-    , Transformer t Tree (MaybeT m)
+    , Transformer t TreeT (MaybeT m)
     )
 
-  distribute =
-    distributeGen
+  distributeT =
+    distributeGenT
 
 instance PrimMonad m => PrimMonad (GenT m) where
   type PrimState (GenT m) =
     PrimState m
+
   primitive =
-    Trans.lift . primitive
+    lift . primitive
 
 instance MonadIO m => MonadIO (GenT m) where
   liftIO =
-    Trans.lift . liftIO
+    lift . liftIO
 
 instance MonadBase b m => MonadBase b (GenT m) where
   liftBase =
-    Trans.lift . liftBase
+    lift . liftBase
+
+#if __GLASGOW_HASKELL__ >= 806
+deriving via (ReaderT Size (ReaderT Seed (TreeT (MaybeT m))))
+  instance MonadBaseControl b m => MonadBaseControl b (GenT m)
+#else
+instance MonadBaseControl b m => MonadBaseControl b (GenT m) where
+  type StM (GenT m) a = StM (GloopT m) a
+  liftBaseWith g = gloopToGen $ liftBaseWith $ \q -> g (\gen -> q (genToGloop gen))
+  restoreM = gloopToGen . restoreM
+
+type GloopT m = ReaderT Size (ReaderT Seed (TreeT (MaybeT m)))
+
+gloopToGen :: GloopT m a -> GenT m a
+gloopToGen = coerce
+
+genToGloop :: GenT m a -> GloopT m a
+genToGloop = coerce
+#endif
 
 instance MonadThrow m => MonadThrow (GenT m) where
   throwM =
-    Trans.lift . throwM
+    lift . throwM
 
 instance MonadCatch m => MonadCatch (GenT m) where
   catch m onErr =
@@ -647,31 +649,33 @@ instance MonadCatch m => MonadCatch (GenT m) where
 
 instance MonadReader r m => MonadReader r (GenT m) where
   ask =
-    Trans.lift ask
+    lift ask
   local f m =
     mapGenT (local f) m
 
 instance MonadState s m => MonadState s (GenT m) where
   get =
-    Trans.lift get
+    lift get
   put =
-    Trans.lift . put
+    lift . put
   state =
-    Trans.lift . state
+    lift . state
 
 instance MonadWriter w m => MonadWriter w (GenT m) where
   writer =
-    Trans.lift . writer
+    lift . writer
   tell =
-    Trans.lift . tell
+    lift . tell
   listen =
-    mapGenT listen
+    undefined
+    --mapGenT listen
   pass =
-    mapGenT pass
+    undefined
+    --mapGenT pass
 
 instance MonadError e m => MonadError e (GenT m) where
   throwError =
-    Trans.lift . throwError
+    lift . throwError
   catchError m onErr =
     GenT $ \size seed ->
       case Seed.split seed of
@@ -681,22 +685,16 @@ instance MonadError e m => MonadError e (GenT m) where
 
 instance MonadResource m => MonadResource (GenT m) where
   liftResourceT =
-    Trans.lift . liftResourceT
+    lift . liftResourceT
 
 ------------------------------------------------------------------------
 -- Combinators
 
--- | Lift a vanilla 'Gen' in to a 'MonadGen'.
---
-lift :: MonadGen m => Gen a -> m a
-lift =
-  liftGen
-
 -- | Generate a value with no shrinks from a 'Size' and a 'Seed'.
 --
 generate :: MonadGen m => (Size -> Seed -> a) -> m a
 generate f =
-  liftGen . GenT $ \size seed ->
+  fromGenT . GenT $ \size seed ->
     pure (f size seed)
 
 ------------------------------------------------------------------------
@@ -708,14 +706,14 @@ generate f =
 --   the existing shrinks intact.
 --
 shrink :: MonadGen m => (a -> [a]) -> m a -> m a
-shrink =
-  shrinkGen
+shrink f =
+  withGenT $ mapGenT (Tree.expand f)
 
 -- | Throw away a generator's shrink tree.
 --
 prune :: MonadGen m => m a -> m a
 prune =
-  pruneGen
+  withGenT $ mapGenT (Tree.prune 0)
 
 ------------------------------------------------------------------------
 -- Combinators - Size
@@ -736,8 +734,17 @@ resize size gen =
 -- | Adjust the size parameter by transforming it with the given function.
 --
 scale :: MonadGen m => (Size -> Size) -> m a -> m a
-scale =
-  scaleGen
+scale f =
+  withGenT $ \gen ->
+    GenT $ \size0 seed ->
+      let
+        size =
+          f size0
+      in
+        if size < 0 then
+          error "Hedgehog.Gen.scale: negative size"
+        else
+          runGenT size seed gen
 
 -- | Make a generator smaller by scaling its size parameter.
 --
@@ -893,7 +900,7 @@ word64 =
 
 -- | Generates a random floating-point number in the @[inclusive,exclusive)@ range.
 --
---   This generator works the same as 'integral', but for floating point numbers.
+--   /This generator works the same as 'integral', but for floating point numbers./
 --
 realFloat :: (MonadGen m, RealFloat a) => Range a -> m a
 realFloat range =
@@ -957,6 +964,8 @@ enum lo hi =
 -- enumBounded :: 'Gen' 'Bool'
 -- @
 --
+--   /This is implemented in terms of the 'Enum' class, and thus may be/
+--   /partial for integral types larger than 'Int', e.g. 'Word64'./
 enumBounded :: (MonadGen m, Enum a, Bounded a) => m a
 enumBounded =
   enum minBound maxBound
@@ -1047,11 +1056,19 @@ latin1 =
   enum '\0' '\255'
 
 -- | Generates a Unicode character, excluding noncharacters and invalid standalone surrogates:
---   @'\0'..'\1114111' (excluding '\55296'..'\57343')@
+--   @'\0'..'\1114111' (excluding '\55296'..'\57343', '\65534', '\65535')@
 --
-unicode :: MonadGen m => m Char
+unicode :: (MonadGen m) => m Char
 unicode =
-  filter (not . isNoncharacter) $ filter (not . isSurrogate) unicodeAll
+  let
+    s1 =
+      (55296, enum '\0' '\55295')
+    s2 =
+      (8190, enum '\57344' '\65533')
+    s3 =
+      (1048576, enum '\65536' '\1114111')
+  in
+    frequency [s1, s2, s3]
 
 -- | Generates a Unicode character, including noncharacters and invalid standalone surrogates:
 --   @'\0'..'\1114111'@
@@ -1229,7 +1246,7 @@ recursive f nonrec rec =
 --
 discard :: MonadGen m => m a
 discard =
-  liftGen mzero
+  fromGenT empty
 
 -- | Discards the generator if the generated value does not satisfy the
 --   predicate.
@@ -1242,37 +1259,67 @@ ensure p gen = do
   else
     discard
 
+fromPred :: (a -> Bool) -> a -> Maybe a
+fromPred p a = a <$ guard (p a)
+
 -- | Generates a value that satisfies a predicate.
 --
 --   This is essentially:
 --
 -- @
--- filter p gen = 'mfilter' p gen '<|>' filter p gen
+--   filter p gen = 'mfilter' p gen '<|>' filter p gen
 -- @
 --
 --   It differs from the above in that we keep some state to avoid looping
 --   forever. If we trigger these limits then the whole generator is discarded.
 --
-filter :: MonadGen m => (a -> Bool) -> m a -> m a
-filter p gen =
+filter :: (MonadGen m, GenBase m ~ Identity) => (a -> Bool) -> m a -> m a
+filter p =
+  mapMaybe (fromPred p)
+
+mapMaybe :: (MonadGen m, GenBase m ~ Identity) => (a -> Maybe b) -> m a -> m b
+mapMaybe p gen0 =
   let
     try k =
       if k > 100 then
         discard
       else do
-        x <- scale (2 * k +) gen
-        if p x then
-          pure x
-        else
-          try (k + 1)
+        (x, gen) <- freeze $ scale (2 * k +) gen0
+
+        case p x of
+          Just _ ->
+            withGenT (mapGenT (Tree.mapMaybeMaybeT p)) gen
+          Nothing ->
+            try (k + 1)
+  in
+    try 0
+
+filterT :: MonadGen m => (a -> Bool) -> m a -> m a
+filterT p =
+  mapMaybeT (fromPred p)
+
+mapMaybeT :: MonadGen m => (a -> Maybe b) -> m a -> m b
+mapMaybeT p gen0 =
+  let
+    try k =
+      if k > 100 then
+        discard
+      else do
+        (x, gen) <- freeze $ scale (2 * k +) gen0
+
+        case p x of
+          Just _ ->
+            withGenT (mapGenT (Tree.mapMaybeT p)) gen
+          Nothing ->
+            try (k + 1)
   in
     try 0
 
 -- | Runs a 'Maybe' generator until it produces a 'Just'.
 --
---   This is implemented using 'filter' and has the same caveats.
+--   /This is implemented using 'filter' and has the same caveats./
 --
-just :: MonadGen m => m (Maybe a) -> m a
+just :: (MonadGen m, GenBase m ~ Identity) => m (Maybe a) -> m a
 just g = do
   mx <- filter Maybe.isJust g
   case mx of
@@ -1281,6 +1328,19 @@ just g = do
     Nothing ->
       error "Hedgehog.Gen.just: internal error, unexpected Nothing"
 
+-- | Runs a 'Maybe' generator until it produces a 'Just'.
+--
+--   /This is implemented using 'filter' and has the same caveats./
+--
+justT :: MonadGen m => m (Maybe a) -> m a
+justT g = do
+  mx <- filterT Maybe.isJust g
+  case mx of
+    Just x ->
+      pure x
+    Nothing ->
+      error "Hedgehog.Gen.just: internal error, unexpected Nothing"
+
 ------------------------------------------------------------------------
 -- Combinators - Collections
 
@@ -1298,12 +1358,19 @@ maybe gen =
 --
 list :: MonadGen m => Range Int -> m a -> m [a]
 list range gen =
-  sized $ \size ->
-    (traverse snd =<<) .
-    ensure (atLeast $ Range.lowerBound size range) .
-    shrink Shrink.list $ do
-      k <- integral_ range
-      replicateM k (freeze gen)
+  let
+     interleave =
+       (interleaveTreeT . nodeValue =<<)
+  in
+    sized $ \size ->
+      ensure (atLeast $ Range.lowerBound size range) .
+      withGenT (mapGenT (TreeT . interleave . runTreeT)) $ do
+        n <- integral_ range
+        replicateM n (toTreeMaybeT gen)
+
+interleaveTreeT :: Monad m => [TreeT m a] -> m (NodeT m [a])
+interleaveTreeT =
+  fmap Tree.interleave . traverse runTreeT
 
 -- | Generates a seq using a 'Range' to determine the length.
 --
@@ -1416,7 +1483,14 @@ deriving instance Traversable (Vec n)
 --
 freeze :: MonadGen m => m a -> m (a, m a)
 freeze =
-  freezeGen
+  withGenT $ \gen ->
+    GenT $ \size seed -> do
+      mx <- lift . lift . runMaybeT . runTreeT $ runGenT size seed gen
+      case mx of
+        Nothing ->
+          empty
+        Just (NodeT x xs) ->
+          pure (x, fromGenT . fromTreeMaybeT . Tree.fromNodeT $ NodeT x xs)
 
 shrinkSubterms :: Subterms n a -> [Subterms n a]
 shrinkSubterms = \case
@@ -1512,20 +1586,47 @@ subsequence xs =
 
 -- | Generates a random permutation of a list.
 --
---   This shrinks towards the order of the list being identical to the input
---   list.
+--   /This shrinks towards the order of the list being identical to the input/
+--   /list./
 --
 shuffle :: MonadGen m => [a] -> m [a]
-shuffle = \case
-  [] ->
-    pure []
-  xs0 -> do
-    n <- integral $ Range.constant 0 (length xs0 - 1)
-    case splitAt n xs0 of
-      (xs, y : ys) ->
-        (y :) <$> shuffle (xs ++ ys)
-      (_, []) ->
-        error "Hedgehog.Gen.shuffle: internal error, split generated empty list"
+-- We shuffle sequences instead of lists to make extracting an arbitrary
+-- element logarithmic instead of linear, and to make length calculation
+-- constant-time instead of linear. We could probably do better, but
+-- this is at least reasonably quick.
+shuffle = fmap toList . shuffleSeq . Seq.fromList
+
+-- | Generates a random permutation of a sequence.
+--
+--   /This shrinks towards the order of the sequence being identical to the input/
+--   /sequence./
+--
+shuffleSeq :: MonadGen m => Seq a -> m (Seq a)
+shuffleSeq xs =
+  if null xs then
+    pure Seq.empty
+  else do
+    n <- integral $ Range.constant 0 (length xs - 1)
+#if MIN_VERSION_containers(0,5,8)
+    -- Data.Sequence should offer a version of deleteAt that returns the
+    -- deleted element, but it does not currently do so. Lookup followed
+    -- by deletion seems likely faster than splitting and then appending,
+    -- but I haven't actually tested that. It's certainly easier to see
+    -- what's going on.
+    case Seq.lookup n xs of
+      Just y ->
+        (y Seq.<|) <$> shuffleSeq (Seq.deleteAt n xs)
+      Nothing ->
+        error "Hedgehog.Gen.shuffleSeq: internal error, lookup in empty sequence"
+#else
+    case Seq.splitAt n xs of
+      (beginning, end) ->
+        case Seq.viewl end of
+          y Seq.:< end' ->
+            (y Seq.<|) <$> shuffleSeq (beginning Seq.>< end')
+          Seq.EmptyL ->
+            error "Hedgehog.Gen.shuffleSeq: internal error, lookup in empty sequence"
+#endif
 
 ------------------------------------------------------------------------
 -- Sampling
@@ -1541,46 +1642,14 @@ sample gen =
           error "Hedgehog.Gen.sample: too many discards, could not generate a sample"
         else do
           seed <- Seed.random
-          case runIdentity . runMaybeT . runTree $ runGenT 30 seed gen of
+          case evalGen 30 seed gen of
             Nothing ->
               loop (n - 1)
             Just x ->
-              pure $ nodeValue x
+              pure $ Tree.treeValue x
     in
       loop (100 :: Int)
 
--- | Print the value produced by a generator, and the first level of shrinks,
---   for the given size and seed.
---
---   Use 'print' to generate a value from a random seed.
---
-printWith :: (MonadIO m, Show a) => Size -> Seed -> Gen a -> m ()
-printWith size seed gen =
-  liftIO $ do
-    let
-      Node x ss =
-        runIdentity . runTree $ renderNodes size seed gen
-
-    putStrLn "=== Outcome ==="
-    putStrLn x
-    putStrLn "=== Shrinks ==="
-
-    for_ ss $ \s ->
-      let
-        Node y _ =
-          runIdentity $ runTree s
-      in
-        putStrLn y
-
--- | Print the shrink tree produced by a generator, for the given size and
---   seed.
---
---   Use 'printTree' to generate a value from a random seed.
---
-printTreeWith :: (MonadIO m, Show a) => Size -> Seed -> Gen a -> m ()
-printTreeWith size seed gen = do
-  liftIO . putStr . runIdentity . Tree.render $ renderNodes size seed gen
-
 -- | Run a generator with a random seed and print the outcome, and the first
 --   level of shrinks.
 --
@@ -1600,6 +1669,35 @@ print gen = do
   seed <- liftIO Seed.random
   printWith 30 seed gen
 
+-- | Print the value produced by a generator, and the first level of shrinks,
+--   for the given size and seed.
+--
+--   Use 'print' to generate a value from a random seed.
+--
+printWith :: (MonadIO m, Show a) => Size -> Seed -> Gen a -> m ()
+printWith size seed gen =
+  liftIO $ do
+    case evalGen size seed gen of
+      Nothing -> do
+        putStrLn "=== Outcome ==="
+        putStrLn "<discard>"
+
+      Just tree_ -> do
+        let
+          NodeT x ss =
+            runIdentity (runTreeT tree_)
+
+        putStrLn "=== Outcome ==="
+        putStrLn (show x)
+        putStrLn "=== Shrinks ==="
+
+        for_ ss $ \s ->
+          let
+            NodeT y _ =
+              runIdentity $ runTreeT s
+          in
+            putStrLn (show y)
+
 -- | Run a generator with a random seed and print the resulting shrink tree.
 --
 -- @
@@ -1622,11 +1720,26 @@ printTree gen = do
   seed <- liftIO Seed.random
   printTreeWith 30 seed gen
 
--- | Render a generator as a tree of strings.
+-- | Print the shrink tree produced by a generator, for the given size and
+--   seed.
 --
-renderNodes :: (Monad m, Show a) => Size -> Seed -> Gen a -> Tree m String
-renderNodes size seed =
-  fmap (Maybe.maybe "<discard>" show) . runDiscardEffect . runGenT size seed . lift
+--   Use 'printTree' to generate a value from a random seed.
+--
+printTreeWith :: (MonadIO m, Show a) => Size -> Seed -> Gen a -> m ()
+printTreeWith size seed gen = do
+  liftIO . putStr $
+    renderTree size seed gen
+
+-- | Render the shrink tree produced by a generator, for the given size and
+--   seed.
+--
+renderTree :: Show a => Size -> Seed -> Gen a -> String
+renderTree size seed gen =
+  case evalGen size seed gen of
+    Nothing ->
+      "<discard>"
+    Just x ->
+      Tree.render (fmap show x)
 
 ------------------------------------------------------------------------
 -- Internal
