@@ -1,6 +1,11 @@
 {-# OPTIONS_HADDOCK not-home #-}
+{-# LANGUAGE ApplicativeDo #-}
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFoldable #-}
 {-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveLift #-}
+{-# LANGUAGE DeriveTraversable #-}
 {-# LANGUAGE DoAndIfThenElse #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
@@ -8,6 +13,7 @@
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TemplateHaskell #-}
@@ -21,8 +27,11 @@ module Hedgehog.Internal.Property (
   , PropertyName(..)
   , PropertyConfig(..)
   , TestLimit(..)
+  , TestCount(..)
   , DiscardLimit(..)
+  , DiscardCount(..)
   , ShrinkLimit(..)
+  , ShrinkCount(..)
   , ShrinkRetries(..)
   , withTests
   , withDiscards
@@ -34,17 +43,20 @@ module Hedgehog.Internal.Property (
   , forAllT
   , forAllWith
   , forAllWithT
+  , defaultMinTests
   , discard
 
   -- * Group
   , Group(..)
   , GroupName(..)
+  , PropertyCount(..)
 
   -- * TestT
   , MonadTest(..)
   , Test
   , TestT(..)
   , Log(..)
+  , Journal(..)
   , Failure(..)
   , Diff(..)
   , annotate
@@ -54,6 +66,7 @@ module Hedgehog.Internal.Property (
   , failure
   , success
   , assert
+  , diff
   , (===)
   , (/==)
 
@@ -63,6 +76,34 @@ module Hedgehog.Internal.Property (
   , evalEither
   , evalExceptT
 
+  -- * Coverage
+  , Coverage(..)
+  , Label(..)
+  , LabelName(..)
+  , cover
+  , classify
+  , label
+  , collect
+  , coverPercentage
+  , labelCovered
+  , coverageSuccess
+  , coverageFailures
+  , journalCoverage
+
+  , Cover(..)
+  , CoverCount(..)
+  , CoverPercentage(..)
+  , toCoverCount
+
+  -- * Confidence
+  , Confidence(..)
+  , TerminationCriteria(..)
+  , confidenceSuccess
+  , confidenceFailure
+  , withConfidence
+  , verifiedTermination
+  , defaultConfidence
+
   -- * Internal
   -- $internal
   , defaultConfig
@@ -76,6 +117,8 @@ module Hedgehog.Internal.Property (
   , mkTestT
   , runTest
   , runTestT
+
+  , wilsonBounds
   ) where
 
 import           Control.Applicative (Alternative(..))
@@ -84,6 +127,8 @@ import           Control.Monad.Base (MonadBase(..))
 import           Control.Monad.Catch (MonadThrow(..), MonadCatch(..))
 import           Control.Monad.Catch (SomeException(..), displayException)
 import           Control.Monad.Error.Class (MonadError(..))
+import           Control.Monad.Fail (MonadFail (..))
+import qualified Control.Monad.Fail as Fail
 import           Control.Monad.IO.Class (MonadIO(..))
 import           Control.Monad.Morph (MFunctor(..))
 import           Control.Monad.Primitive (PrimMonad(..))
@@ -108,9 +153,14 @@ import qualified Control.Monad.Trans.Writer.Strict as Strict
 
 import qualified Data.Char as Char
 import           Data.Functor.Identity (Identity(..))
+import           Data.Int (Int64)
+import           Data.Map (Map)
+import qualified Data.Map.Strict as Map
+import           Data.Number.Erf (invnormcdf)
 import qualified Data.List as List
-import           Data.Semigroup (Semigroup)
+import           Data.Semigroup (Semigroup(..))
 import           Data.String (IsString)
+import           Data.Ratio ((%))
 import           Data.Typeable (typeOf)
 
 import           Hedgehog.Internal.Distributive
@@ -120,7 +170,7 @@ import qualified Hedgehog.Internal.Gen as Gen
 import           Hedgehog.Internal.Show
 import           Hedgehog.Internal.Source
 
-import           Language.Haskell.TH.Lift (deriveLift)
+import           Language.Haskell.TH.Syntax (Lift)
 
 ------------------------------------------------------------------------
 
@@ -151,6 +201,18 @@ newtype PropertyT m a =
     , MonadState s
     , MonadError e
     )
+-- NOTE: Move this to the deriving list above when we drop 7.10
+deriving instance MonadResource m => MonadResource (PropertyT m)
+
+-- NOTE: Move this to the deriving list above when we drop 8.0
+#if __GLASGOW_HASKELL__ >= 802
+deriving instance MonadBaseControl b m => MonadBaseControl b (PropertyT m)
+#else
+instance MonadBaseControl b m => MonadBaseControl b (PropertyT m) where
+  type StM (PropertyT m) a = StM (TestT (GenT m)) a
+  liftBaseWith f = PropertyT $ liftBaseWith $ \rib -> f (rib . unPropertyT)
+  restoreM = PropertyT . restoreM
+#endif
 
 -- | A test monad allows the assertion of expectations.
 --
@@ -161,7 +223,7 @@ type Test =
 --
 newtype TestT m a =
   TestT {
-      unTest :: ExceptT Failure (Lazy.WriterT [Log] m) a
+      unTest :: ExceptT Failure (Lazy.WriterT Journal m) a
     } deriving (
       Functor
     , Applicative
@@ -175,7 +237,7 @@ newtype TestT m a =
 
 -- | The name of a property.
 --
---   Can be constructed using `OverloadedStrings`:
+--   Should be constructed using `OverloadedStrings`:
 --
 -- @
 --   "apples" :: PropertyName
@@ -184,17 +246,26 @@ newtype TestT m a =
 newtype PropertyName =
   PropertyName {
       unPropertyName :: String
-    } deriving (Eq, Ord, Show, IsString, Semigroup)
+    } deriving (Eq, Ord, Show, IsString, Semigroup, Lift)
+
+-- | The acceptable occurrence of false positives
+--
+--   Example, @Confidence 10^9@ would mean that you'd accept a false positive
+--   for 1 in 10^9 tests.
+newtype Confidence =
+  Confidence {
+    unConfidence :: Int64
+  } deriving (Eq, Ord, Show, Num, Lift)
 
 -- | Configuration for a property test.
 --
 data PropertyConfig =
   PropertyConfig {
-      propertyTestLimit :: !TestLimit
-    , propertyDiscardLimit :: !DiscardLimit
+      propertyDiscardLimit :: !DiscardLimit
     , propertyShrinkLimit :: !ShrinkLimit
     , propertyShrinkRetries :: !ShrinkRetries
-    } deriving (Eq, Ord, Show)
+    , propertyTerminationCriteria :: !TerminationCriteria
+    } deriving (Eq, Ord, Show, Lift)
 
 -- | The number of successful tests that need to be run before a property test
 --   is considered successful.
@@ -207,6 +278,18 @@ data PropertyConfig =
 --
 newtype TestLimit =
   TestLimit Int
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Lift)
+
+-- | The number of tests a property ran successfully.
+--
+newtype TestCount =
+  TestCount Int
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
+
+-- | The number of tests a property had to discard.
+--
+newtype DiscardCount =
+  DiscardCount Int
   deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
 
 -- | The number of discards to allow before giving up.
@@ -220,7 +303,7 @@ newtype TestLimit =
 --
 newtype DiscardLimit =
   DiscardLimit Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Lift)
 
 -- | The number of shrinks to try before giving up on shrinking.
 --
@@ -232,6 +315,12 @@ newtype DiscardLimit =
 --
 newtype ShrinkLimit =
   ShrinkLimit Int
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Lift)
+
+-- | The numbers of times a property was able to shrink after a failing test.
+--
+newtype ShrinkCount =
+  ShrinkCount Int
   deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
 
 -- | The number of times to re-run a test during shrinking. This is useful if
@@ -251,7 +340,7 @@ newtype ShrinkLimit =
 --
 newtype ShrinkRetries =
   ShrinkRetries Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Lift)
 
 -- | A named collection of property tests.
 --
@@ -263,7 +352,7 @@ data Group =
 
 -- | The name of a group of properties.
 --
---   Can be constructed using `OverloadedStrings`:
+--   Should be constructed using `OverloadedStrings`:
 --
 -- @
 --   "fruit" :: GroupName
@@ -272,12 +361,24 @@ data Group =
 newtype GroupName =
   GroupName {
       unGroupName :: String
-    } deriving (Eq, Ord, Show, IsString, Semigroup)
+    } deriving (Eq, Ord, Show, IsString, Semigroup, Lift)
+
+-- | The number of properties in a group.
+--
+newtype PropertyCount =
+  PropertyCount Int
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
+
+data TerminationCriteria =
+    EarlyTermination Confidence TestLimit
+  | NoEarlyTermination Confidence TestLimit
+  | NoConfidenceTermination TestLimit
+  deriving (Eq, Ord, Show, Lift)
 
 --
 -- FIXME This whole Log/Failure thing could be a lot more structured to allow
 -- FIXME for richer user controlled error messages, think Doc. Ideally we'd
--- FIXME allow user's to create their own diffs anywhere.
+-- FIXME allow user's to crete their own diffs anywhere.
 --
 
 -- | Log messages which are recorded during a test run.
@@ -285,8 +386,15 @@ newtype GroupName =
 data Log =
     Annotation (Maybe Span) String
   | Footnote String
+  | Label (Label Cover)
     deriving (Eq, Show)
 
+-- | A record containing the details of a test run.
+newtype Journal =
+  Journal {
+      journalLogs :: [Log]
+    } deriving (Eq, Show, Semigroup, Monoid)
+
 -- | Details on where and why a test failed.
 --
 data Failure =
@@ -305,18 +413,89 @@ data Diff =
     , diffValue :: ValueDiff
     } deriving (Eq, Show)
 
+-- | Whether a test is covered by a classifier, and therefore belongs to a
+--   'Class'.
+--
+data Cover =
+    NoCover
+  | Cover
+    deriving (Eq, Ord, Show)
+
+-- | The total number of tests which are covered by a classifier.
+--
+--   Can be constructed using numeric literals:
+--
+-- @
+--   30 :: CoverCount
+-- @
+--
+newtype CoverCount =
+  CoverCount {
+      unCoverCount :: Int
+    } deriving (Eq, Ord, Show, Num)
+
+-- | The relative number of tests which are covered by a classifier.
+--
+--   Can be constructed using numeric literals:
+--
+-- @
+--   30 :: CoverPercentage
+-- @
+--
+newtype CoverPercentage =
+  CoverPercentage {
+      unCoverPercentage :: Double
+    } deriving (Eq, Ord, Show, Num, Fractional)
+
+-- | The name of a classifier.
+--
+--   Should be constructed using `OverloadedStrings`:
+--
+-- @
+--   "apples" :: LabelName
+-- @
+--
+newtype LabelName =
+  LabelName {
+      unLabelName :: String
+    } deriving (Eq, Monoid, Ord, Semigroup, Show, IsString)
+
+-- | The extent to which a test is covered by a classifier.
+--
+--   /When a classifier's coverage does not exceed the required minimum, the/
+--   /test will be failed./
+--
+data Label a =
+  MkLabel {
+      labelName :: !LabelName
+    , labelLocation :: !(Maybe Span)
+    , labelMinimum :: !CoverPercentage
+    , labelAnnotation :: !a
+    } deriving (Eq, Show, Functor, Foldable, Traversable)
+
+-- | The extent to which all classifiers cover a test.
+--
+--   /When a given classification's coverage does not exceed the required/
+--   /minimum, the test will be failed./
+--
+newtype Coverage a =
+  Coverage {
+      coverageLabels :: Map LabelName (Label a)
+    } deriving (Eq, Show, Functor, Foldable, Traversable)
+
 ------------------------------------------------------------------------
 -- TestT
 
 instance Monad m => Monad (TestT m) where
   return =
-    TestT . return
+    pure
 
   (>>=) m k =
     TestT $
       unTest m >>=
       unTest . k
 
+instance Monad m => MonadFail (TestT m) where
   fail err =
     TestT . ExceptT . pure . Left $ Failure Nothing err Nothing
 
@@ -328,16 +507,16 @@ instance MFunctor TestT where
   hoist f =
     TestT . hoist (hoist f) . unTest
 
-instance Distributive TestT where
+instance MonadTransDistributive TestT where
   type Transformer t TestT m = (
-      Transformer t (Lazy.WriterT [Log]) m
-    , Transformer t (ExceptT Failure) (Lazy.WriterT [Log] m)
+      Transformer t (Lazy.WriterT Journal) m
+    , Transformer t (ExceptT Failure) (Lazy.WriterT Journal m)
     )
 
-  distribute =
+  distributeT =
     hoist TestT .
-    distribute .
-    hoist distribute .
+    distributeT .
+    hoist distributeT .
     unTest
 
 instance PrimMonad m => PrimMonad (TestT m) where
@@ -362,10 +541,10 @@ instance MonadResource m => MonadResource (TestT m) where
 
 instance MonadTransControl TestT where
   type StT TestT a =
-    (Either Failure a, [Log])
+    (Either Failure a, Journal)
 
   liftWith f =
-    mkTestT . fmap (, []) . fmap Right $ f $ runTestT
+    mkTestT . fmap (, mempty) . fmap Right $ f $ runTestT
 
   restoreT =
     mkTestT
@@ -435,19 +614,19 @@ instance MonadTest m => MonadTest (ResourceT m) where
   liftTest =
     lift . liftTest
 
-mkTestT :: m (Either Failure a, [Log]) -> TestT m a
+mkTestT :: m (Either Failure a, Journal) -> TestT m a
 mkTestT =
   TestT . ExceptT . Lazy.WriterT
 
-mkTest :: (Either Failure a, [Log]) -> Test a
+mkTest :: (Either Failure a, Journal) -> Test a
 mkTest =
   mkTestT . Identity
 
-runTestT :: TestT m a -> m (Either Failure a, [Log])
+runTestT :: TestT m a -> m (Either Failure a, Journal)
 runTestT =
   Lazy.runWriterT . runExceptT . unTest
 
-runTest :: Test a -> (Either Failure a, [Log])
+runTest :: Test a -> (Either Failure a, Journal)
 runTest =
   runIdentity . runTestT
 
@@ -455,14 +634,14 @@ runTest =
 --
 writeLog :: MonadTest m => Log -> m ()
 writeLog x =
-  liftTest $ mkTest (pure (), [x])
+  liftTest $ mkTest (pure (), (Journal [x]))
 
 -- | Fail the test with an error message, useful for building other failure
 --   combinators.
 --
 failWith :: (MonadTest m, HasCallStack) => Maybe Diff -> String -> m a
-failWith diff msg =
-  liftTest $ mkTest (Left $ Failure (getCaller callStack) msg diff, [])
+failWith mdiff msg =
+  liftTest $ mkTest (Left $ Failure (getCaller callStack) msg mdiff, mempty)
 
 -- | Annotates the source code with a message that might be useful for
 --   debugging a test failure.
@@ -492,21 +671,30 @@ footnoteShow :: (MonadTest m, Show a) => a -> m ()
 footnoteShow =
   writeLog . Footnote . showPretty
 
--- | Fails with an error which shows the difference between two values.
---
+-- | Fails with an error that shows the difference between two values.
 failDiff :: (MonadTest m, Show a, Show b, HasCallStack) => a -> b -> m ()
 failDiff x y =
   case valueDiff <$> mkValue x <*> mkValue y of
     Nothing ->
       withFrozenCallStack $
-        failWith Nothing $ unlines [
-            "━━━ Not Equal ━━━"
+        failWith Nothing $
+        unlines $ [
+            "Failed"
+          , "━━ lhs ━━"
           , showPretty x
+          , "━━ rhs ━━"
           , showPretty y
           ]
-    Just diff ->
+
+    Just vdiff@(ValueSame _) ->
       withFrozenCallStack $
-        failWith (Just $ Diff "Failed (" "- lhs" "=/=" "+ rhs" ")" diff) ""
+        failWith (Just $
+          Diff "━━━ Failed ("  "" "no differences" "" ") ━━━" vdiff) ""
+
+    Just vdiff ->
+      withFrozenCallStack $
+        failWith (Just $
+          Diff "━━━ Failed (" "- lhs" ") (" "+ rhs" ") ━━━" vdiff) ""
 
 -- | Fails with an error which renders the type of an exception and its error
 --   message.
@@ -515,7 +703,7 @@ failException :: (MonadTest m, HasCallStack) => SomeException -> m a
 failException (SomeException x) =
   withFrozenCallStack $
     failWith Nothing $ unlines [
-        "━━━ Exception: " ++ show (typeOf x) ++ " ━━━"
+        "━━━ Exception (" ++ show (typeOf x) ++ ") ━━━"
       , List.dropWhileEnd Char.isSpace (displayException x)
       ]
 
@@ -541,33 +729,48 @@ assert b = do
   else
     withFrozenCallStack failure
 
+-- | Fails the test and shows a git-like diff if the comparison operation
+--   evaluates to 'False' when applied to its arguments.
+--
+--   The comparison function is the second argument, which may be
+--   counter-intuitive to Haskell programmers. However, it allows operators to
+--   be written infix for easy reading:
+--
+-- @
+--   diff y (<) 87
+--   diff x (<=) 'r'
+-- @
+--
+--   /This function behaves like the unix `diff` tool, which gives a `0` exit/
+--   /code if the compared files are identical, or a `1` exit code code/
+--   /otherwise. Like unix `diff`, if the arguments fail the comparison, a diff/
+--   /is shown./
+--
+diff :: (MonadTest m, Show a, Show b, HasCallStack) => a -> (a -> b -> Bool) -> b -> m ()
+diff x op y = do
+  ok <- withFrozenCallStack $ eval (x `op` y)
+  if ok then
+    success
+  else
+    withFrozenCallStack $ failDiff x y
+
 infix 4 ===
 
 -- | Fails the test if the two arguments provided are not equal.
 --
 (===) :: (MonadTest m, Eq a, Show a, HasCallStack) => a -> a -> m ()
-(===) x y = do
-  ok <- withFrozenCallStack $ eval (x == y)
-  if ok then
-    success
-  else
-    withFrozenCallStack $ failDiff x y
+(===) x y =
+  withFrozenCallStack $
+    diff x (==) y
 
 infix 4 /==
 
 -- | Fails the test if the two arguments provided are equal.
 --
 (/==) :: (MonadTest m, Eq a, Show a, HasCallStack) => a -> a -> m ()
-(/==) x y = do
-  ok <- withFrozenCallStack $ eval (x /= y)
-  if ok then
-    success
-  else
-    withFrozenCallStack $
-      failWith Nothing $ unlines [
-          "━━━ Both equal to ━━━"
-        , showPretty x
-        ]
+(/==) x y =
+  withFrozenCallStack $
+    diff x (/=) y
 
 -- | Fails the test if the value throws an exception when evaluated to weak
 --   head normal form (WHNF).
@@ -618,20 +821,24 @@ instance MonadTrans PropertyT where
   lift =
     PropertyT . lift . lift
 
+instance Monad m => MonadFail (PropertyT m) where
+  fail err =
+    PropertyT (Fail.fail err)
+
 instance MFunctor PropertyT where
   hoist f =
     PropertyT . hoist (hoist f) . unPropertyT
 
-instance Distributive PropertyT where
+instance MonadTransDistributive PropertyT where
   type Transformer t PropertyT m = (
       Transformer t GenT m
     , Transformer t TestT (GenT m)
     )
 
-  distribute =
+  distributeT =
     hoist PropertyT .
-    distribute .
-    hoist distribute .
+    distributeT .
+    hoist distributeT .
     unPropertyT
 
 instance PrimMonad m => PrimMonad (PropertyT m) where
@@ -680,10 +887,11 @@ forAllWithT render gen = do
 --
 forAllWith :: (Monad m, HasCallStack) => (a -> String) -> Gen a -> PropertyT m a
 forAllWith render gen =
-  withFrozenCallStack $ forAllWithT render $ Gen.lift gen
+  withFrozenCallStack $ forAllWithT render $ Gen.generalize gen
 
 -- | Generates a random input for the test by running the provided generator.
 --
+--
 forAllT :: (Monad m, Show a, HasCallStack) => GenT m a -> PropertyT m a
 forAllT gen =
   withFrozenCallStack $ forAllWithT showPretty gen
@@ -698,7 +906,7 @@ forAll gen =
 --
 discard :: Monad m => PropertyT m a
 discard =
-  PropertyT $ lift Gen.discard
+  PropertyT $ lift (Gen.generalize Gen.discard)
 
 -- | Lift a test in to a property.
 --
@@ -708,14 +916,9 @@ discard =
 --   instance at the expense of not being able to generate additional inputs
 --   using 'forAll'.
 --
---   One use case for this is writing tests which use 'ResourceT':
---
--- @
---   property $ do
---     n <- forAll $ Gen.int64 Range.linearBounded
---     test . runResourceT $ do
---       -- test with resource usage here
--- @
+--   An example where this is useful is parallel state machine testing, as
+--   'Hedgehog.Internal.State.executeParallel' requires 'MonadBaseControl' 'IO'
+--   in order to be able to spawn threads in 'MonadTest'.
 --
 test :: Monad m => TestT m a -> PropertyT m a
 test =
@@ -729,22 +932,60 @@ test =
 defaultConfig :: PropertyConfig
 defaultConfig =
   PropertyConfig {
-      propertyTestLimit =
-        100
-    , propertyDiscardLimit =
+      propertyDiscardLimit =
         100
     , propertyShrinkLimit =
         1000
     , propertyShrinkRetries =
         0
+    , propertyTerminationCriteria =
+        NoConfidenceTermination defaultMinTests
     }
 
+-- | The minimum amount of tests to run for a 'Property'
+--
+defaultMinTests :: TestLimit
+defaultMinTests = 100
+
+-- | The default confidence allows one false positive in 10^9 tests
+--
+defaultConfidence :: Confidence
+defaultConfidence = 10 ^ (9 :: Int)
+
 -- | Map a config modification function over a property.
 --
 mapConfig :: (PropertyConfig -> PropertyConfig) -> Property -> Property
 mapConfig f (Property cfg t) =
   Property (f cfg) t
 
+-- | Make sure that the result is statistically significant in accordance to
+--   the passed 'Confidence'
+--
+withConfidence :: Confidence -> Property -> Property
+withConfidence c =
+  let
+    setConfidence = \case
+      NoEarlyTermination _ tests -> NoEarlyTermination c tests
+      NoConfidenceTermination tests -> NoEarlyTermination c tests
+      EarlyTermination _ tests -> EarlyTermination c tests
+  in
+    mapConfig $ \config@PropertyConfig{..} ->
+      config
+        { propertyTerminationCriteria =
+            setConfidence propertyTerminationCriteria
+        }
+
+verifiedTermination :: Property -> Property
+verifiedTermination =
+  mapConfig $ \config@PropertyConfig{..} ->
+    let
+      newTerminationCriteria = case propertyTerminationCriteria of
+        NoEarlyTermination c tests -> EarlyTermination c tests
+        NoConfidenceTermination tests -> EarlyTermination defaultConfidence tests
+        EarlyTermination c tests -> EarlyTermination c tests
+    in
+      config { propertyTerminationCriteria = newTerminationCriteria }
+
 -- | Set the number of times a property should be executed before it is considered
 --   successful.
 --
@@ -754,7 +995,14 @@ mapConfig f (Property cfg t) =
 --
 withTests :: TestLimit -> Property -> Property
 withTests n =
-  mapConfig $ \config -> config { propertyTestLimit = n }
+  let
+    setTestLimit tests = \case
+      NoEarlyTermination c _ -> NoEarlyTermination c tests
+      NoConfidenceTermination _ -> NoConfidenceTermination tests
+      EarlyTermination c _ -> EarlyTermination c tests
+  in
+    mapConfig $ \config@PropertyConfig{..} ->
+      config { propertyTerminationCriteria = setTestLimit n propertyTerminationCriteria }
 
 -- | Set the number of times a property is allowed to discard before the test
 --   runner gives up.
@@ -786,15 +1034,219 @@ property m =
     withFrozenCallStack (evalM m)
 
 ------------------------------------------------------------------------
--- FIXME Replace with DeriveLift when we drop 7.10 support.
+-- Coverage
 
-$(deriveLift ''GroupName)
-$(deriveLift ''PropertyName)
-$(deriveLift ''PropertyConfig)
-$(deriveLift ''TestLimit)
-$(deriveLift ''DiscardLimit)
-$(deriveLift ''ShrinkLimit)
-$(deriveLift ''ShrinkRetries)
+instance Semigroup Cover where
+  (<>) NoCover NoCover =
+    NoCover
+  (<>) _ _ =
+    Cover
+
+instance Monoid Cover where
+  mempty =
+    NoCover
+  mappend =
+    (<>)
+
+instance Semigroup CoverCount where
+  (<>) (CoverCount n0) (CoverCount n1) =
+    CoverCount (n0 + n1)
+
+instance Monoid CoverCount where
+  mempty =
+    CoverCount 0
+  mappend =
+    (<>)
+
+toCoverCount :: Cover -> CoverCount
+toCoverCount = \case
+  NoCover ->
+    CoverCount 0
+  Cover ->
+    CoverCount 1
+
+-- | This semigroup is right biased. The name, location and percentage from the
+--   rightmost `Label` will be kept. This shouldn't be a problem since the
+--   library doesn't allow setting multiple classes with the same 'ClassifierName'.
+instance Semigroup a => Semigroup (Label a) where
+  (<>) (MkLabel _ _ _ m0) (MkLabel name location percentage m1) =
+    MkLabel name location percentage (m0 <> m1)
+
+instance Semigroup a => Semigroup (Coverage a) where
+  (<>) (Coverage c0) (Coverage c1) =
+    Coverage $
+      Map.foldrWithKey (Map.insertWith (<>)) c0 c1
+
+instance (Semigroup a, Monoid a) => Monoid (Coverage a) where
+  mempty =
+    Coverage mempty
+  mappend =
+    (<>)
+
+coverPercentage :: TestCount -> CoverCount -> CoverPercentage
+coverPercentage (TestCount tests) (CoverCount count) =
+  let
+    percentage :: Double
+    percentage =
+      fromIntegral count / fromIntegral tests * 100
+
+    thousandths :: Int
+    thousandths =
+      round $ percentage * 10
+  in
+    CoverPercentage (fromIntegral thousandths / 10)
+
+labelCovered :: TestCount -> Label CoverCount -> Bool
+labelCovered tests (MkLabel _ _ minimum_ population) =
+  coverPercentage tests population >= minimum_
+
+-- | All labels are covered
+coverageSuccess :: TestCount -> Coverage CoverCount -> Bool
+coverageSuccess tests =
+  null . coverageFailures tests
+
+coverageFailures :: TestCount -> Coverage CoverCount -> [Label CoverCount]
+coverageFailures tests (Coverage kvs) =
+  filter (not . labelCovered tests) (Map.elems kvs)
+
+-- | Is true when the test coverage satisfies the specified 'Confidence'
+--   contstraint for all 'Coverage CoverCount's
+confidenceSuccess :: TestCount -> Confidence -> Coverage CoverCount -> Bool
+confidenceSuccess tests confidence =
+  let
+    assertLow :: Label CoverCount -> Bool
+    assertLow coverCount@MkLabel{..} =
+      fst (boundsForLabel tests confidence coverCount)
+        >= unCoverPercentage labelMinimum / 100.0
+  in
+    and . fmap assertLow . Map.elems . coverageLabels
+
+-- | Is true when there exists a label that is sure to have failed according to
+--   the 'Confidence' constraint
+confidenceFailure :: TestCount -> Confidence -> Coverage CoverCount -> Bool
+confidenceFailure tests confidence =
+  let
+    assertHigh :: Label CoverCount -> Bool
+    assertHigh coverCount@MkLabel{..} =
+      snd (boundsForLabel tests confidence coverCount)
+        < (unCoverPercentage labelMinimum / 100.0)
+  in
+    or . fmap assertHigh . Map.elems . coverageLabels
+
+boundsForLabel :: TestCount -> Confidence -> Label CoverCount -> (Double, Double)
+boundsForLabel tests confidence MkLabel{..} =
+  wilsonBounds
+    (fromIntegral $ unCoverCount labelAnnotation)
+    (fromIntegral tests)
+    (1 / fromIntegral (unConfidence confidence))
+
+-- In order to get an accurate measurement with small sample sizes, we're
+-- using the Wilson score interval
+-- (<https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval#Wilson_score_interval
+-- wikipedia>) instead of a normal approximation interval.
+wilsonBounds :: Integer -> Integer -> Double -> (Double, Double)
+wilsonBounds positives count acceptance =
+  let
+    p =
+      fromRational $ positives % count
+    n =
+      fromIntegral count
+    z =
+      invnormcdf $ 1 - acceptance / 2
+
+    midpoint =
+      p + z * z / (2 * n)
+
+    offset =
+      z / (1 + z ** 2 / n) * sqrt (p * (1 - p) / n + z ** 2 / (4 * n ** 2))
+
+    denominator =
+      1 + z * z / n
+
+    low =
+      (midpoint - offset) / denominator
+
+    high =
+      (midpoint + offset) / denominator
+  in
+    (low, high)
+
+fromLabel :: Label a -> Coverage a
+fromLabel x =
+  Coverage $
+    Map.singleton (labelName x) x
+
+unionsCoverage :: Semigroup a => [Coverage a] -> Coverage a
+unionsCoverage =
+  Coverage .
+  Map.unionsWith (<>) .
+  fmap coverageLabels
+
+journalCoverage :: Journal -> Coverage CoverCount
+journalCoverage (Journal logs) =
+  fmap toCoverCount .
+  unionsCoverage $ do
+    Label x <- logs
+    pure (fromLabel x)
+
+-- | Require a certain percentage of the tests to be covered by the
+--   classifier.
+--
+-- @
+--    prop_with_coverage :: Property
+--    prop_with_coverage =
+--      property $ do
+--        match <- forAll Gen.bool
+--        cover 30 "True" $ match
+--        cover 30 "False" $ not match
+-- @
+--
+--   The example above requires a minimum of 30% coverage for both
+--   classifiers. If these requirements are not met, it will fail the test.
+--
+cover :: (MonadTest m, HasCallStack) => CoverPercentage -> LabelName -> Bool -> m ()
+cover minimum_ name covered =
+  let
+    cover_ =
+      if covered then
+        Cover
+      else
+        NoCover
+  in
+    writeLog . Label $
+      MkLabel name (getCaller callStack) minimum_ cover_
+
+-- | Records the proportion of tests which satisfy a given condition.
+--
+-- @
+--    prop_with_classifier :: Property
+--    prop_with_classifier =
+--      property $ do
+--        xs <- forAll $ Gen.list (Range.linear 0 100) Gen.alpha
+--        for_ xs $ \\x -> do
+--          classify "newborns" $ x == 0
+--          classify "children" $ x > 0 && x < 13
+--          classify "teens" $ x > 12 && x < 20
+-- @
+classify :: (MonadTest m, HasCallStack) => LabelName -> Bool -> m ()
+classify name covered =
+  withFrozenCallStack $
+    cover 0 name covered
+
+-- | Add a label for each test run. It produces a table showing the percentage
+--   of test runs that produced each label.
+--
+label :: (MonadTest m, HasCallStack) => LabelName -> m ()
+label name =
+  withFrozenCallStack $
+    cover 0 name True
+
+-- | Like 'label', but uses 'Show' to render its argument for display.
+--
+collect :: (MonadTest m, Show a, HasCallStack) => a -> m ()
+collect x =
+  withFrozenCallStack $
+    cover 0 (LabelName (show x)) True
 
 ------------------------------------------------------------------------
 -- Internal
