@@ -8,6 +8,7 @@
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE TupleSections #-}
+
 module Hedgehog.Internal.Report (
   -- * Report
     Summary(..)
@@ -17,11 +18,6 @@ module Hedgehog.Internal.Report (
   , FailureReport(..)
   , FailedAnnotation(..)
 
-  , ShrinkCount(..)
-  , TestCount(..)
-  , DiscardCount(..)
-  , PropertyCount(..)
-
   , Style(..)
   , Markup(..)
 
@@ -50,11 +46,19 @@ import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.Maybe (mapMaybe, catMaybes)
 import           Data.Semigroup (Semigroup(..))
+import           Data.Traversable (for)
 
 import           Hedgehog.Internal.Config
 import           Hedgehog.Internal.Discovery (Pos(..), Position(..))
 import qualified Hedgehog.Internal.Discovery as Discovery
+import           Hedgehog.Internal.Property (Coverage(..), Label(..), LabelName(..))
+import           Hedgehog.Internal.Property (CoverCount(..), CoverPercentage(..))
 import           Hedgehog.Internal.Property (PropertyName(..), Log(..), Diff(..))
+import           Hedgehog.Internal.Property (ShrinkCount(..), PropertyCount(..))
+import           Hedgehog.Internal.Property (TestCount(..), DiscardCount(..))
+import           Hedgehog.Internal.Property (coverPercentage, coverageFailures)
+import           Hedgehog.Internal.Property (labelCovered)
+
 import           Hedgehog.Internal.Seed (Seed)
 import           Hedgehog.Internal.Show
 import           Hedgehog.Internal.Source
@@ -64,41 +68,18 @@ import           System.Console.ANSI (ColorIntensity(..), Color(..))
 import           System.Console.ANSI (ConsoleLayer(..), ConsoleIntensity(..))
 import           System.Console.ANSI (SGR(..), setSGRCode)
 import           System.Directory (makeRelativeToCurrentDirectory)
+
 #if mingw32_HOST_OS
 import           System.IO (hSetEncoding, stdout, stderr, utf8)
 #endif
 
-import           Text.PrettyPrint.Annotated.WL (Doc, (<+>))
+import           Text.PrettyPrint.Annotated.WL (Doc, (<#>), (<+>))
 import qualified Text.PrettyPrint.Annotated.WL as WL
 import           Text.Printf (printf)
 
 ------------------------------------------------------------------------
 -- Data
 
--- | The numbers of times a property was able to shrink after a failing test.
---
-newtype ShrinkCount =
-  ShrinkCount Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
-
--- | The number of tests a property ran successfully.
---
-newtype TestCount =
-  TestCount Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
-
--- | The number of tests a property had to discard.
---
-newtype DiscardCount =
-  DiscardCount Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
-
--- | The number of properties in a group.
---
-newtype PropertyCount =
-  PropertyCount Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
-
 data FailedAnnotation =
   FailedAnnotation {
       failedSpan :: !(Maybe Span)
@@ -110,6 +91,7 @@ data FailureReport =
       failureSize :: !Size
     , failureSeed :: !Seed
     , failureShrinks :: !ShrinkCount
+    , failureCoverage :: !(Maybe (Coverage CoverCount))
     , failureAnnotations :: ![FailedAnnotation]
     , failureLocation :: !(Maybe Span)
     , failureMessage :: !String
@@ -141,6 +123,7 @@ data Report a =
   Report {
       reportTests :: !TestCount
     , reportDiscards :: !DiscardCount
+    , reportCoverage :: !(Coverage CoverCount)
     , reportStatus :: !a
     } deriving (Show, Functor, Foldable, Traversable)
 
@@ -220,11 +203,14 @@ data Markup =
   | ShrinkingIcon
   | ShrinkingHeader
   | FailedIcon
-  | FailedHeader
+  | FailedText
   | GaveUpIcon
-  | GaveUpHeader
+  | GaveUpText
   | SuccessIcon
-  | SuccessHeader
+  | SuccessText
+  | CoverageIcon
+  | CoverageText
+  | CoverageFill
   | DeclarationLocation
   | StyledLineNo !Style
   | StyledBorder !Style
@@ -279,12 +265,13 @@ mkFailure ::
      Size
   -> Seed
   -> ShrinkCount
+  -> Maybe (Coverage CoverCount)
   -> Maybe Span
   -> String
   -> Maybe Diff
   -> [Log]
   -> FailureReport
-mkFailure size seed shrinks location message diff logs =
+mkFailure size seed shrinks mcoverage location message diff logs =
   let
     inputs =
       mapMaybe takeAnnotation logs
@@ -292,7 +279,7 @@ mkFailure size seed shrinks location message diff logs =
     footnotes =
       mapMaybe takeFootnote logs
   in
-    FailureReport size seed shrinks inputs location message diff footnotes
+    FailureReport size seed shrinks mcoverage inputs location message diff footnotes
 
 ------------------------------------------------------------------------
 -- Pretty Printing
@@ -496,19 +483,19 @@ ppLineDiff = \case
 ppDiff :: Diff -> [Doc Markup]
 ppDiff (Diff prefix removed infix_ added suffix diff) = [
     markup DiffPrefix (WL.text prefix) <>
-    markup DiffRemoved (WL.text removed) <+>
-    markup DiffInfix (WL.text infix_) <+>
+    markup DiffRemoved (WL.text removed) <>
+    markup DiffInfix (WL.text infix_) <>
     markup DiffAdded (WL.text added) <>
     markup DiffSuffix (WL.text suffix)
   ] ++ fmap ppLineDiff (toLineDiff diff)
 
 ppFailureLocation ::
      MonadIO m
-  => String
+  => [Doc Markup]
   -> Maybe Diff
   -> Span
   -> m (Maybe (Declaration (Style, [(Style, Doc Markup)])))
-ppFailureLocation msg mdiff sloc =
+ppFailureLocation msgs mdiff sloc =
   runMaybeT $ do
     decl <- fmap defaultStyle . MaybeT $ readDeclaration sloc
     (startCol, endCol) <- bimap fromIntegral fromIntegral <$> lastLineSpan sloc decl
@@ -523,7 +510,7 @@ ppFailureLocation msg mdiff sloc =
           markup FailureGutter (WL.text "│ ") <> x
 
       msgDocs =
-        fmap ((StyleFailure, ) . ppFailure . markup FailureMessage . WL.text) (List.lines msg)
+        fmap ((StyleFailure, ) . ppFailure . markup FailureMessage) msgs
 
       diffDocs =
         case mdiff of
@@ -619,9 +606,9 @@ ppTextLines :: String -> [Doc Markup]
 ppTextLines =
   fmap WL.text . List.lines
 
-ppFailureReport :: MonadIO m => Maybe PropertyName -> FailureReport -> m (Doc Markup)
-ppFailureReport name (FailureReport size seed _ inputs0 mlocation0 msg mdiff msgs0) = do
-  (msgs, mlocation) <-
+ppFailureReport :: MonadIO m => Maybe PropertyName -> TestCount -> FailureReport -> m [Doc Markup]
+ppFailureReport name tests (FailureReport size seed _ mcoverage inputs0 mlocation0 msg mdiff msgs0) = do
+  (msgs1, mlocation) <-
     case mlocation0 of
       Nothing ->
         -- Move the failure message to the end section if we have
@@ -638,16 +625,37 @@ ppFailureReport name (FailureReport size seed _ inputs0 mlocation0 msg mdiff msg
           pure (docs, Nothing)
 
       Just location0 ->
-        (concatMap ppTextLines msgs0,)
-          <$> ppFailureLocation msg mdiff location0
+        fmap (concatMap ppTextLines msgs0,) $
+          ppFailureLocation (fmap WL.text $ List.lines msg) mdiff location0
 
-  (args, idecls) <- partitionEithers <$> zipWithM ppFailedInput [0..] inputs0
+  coverageLocations <-
+    case mcoverage of
+      Nothing ->
+        pure []
+      Just coverage ->
+        for (coverageFailures tests coverage) $ \(MkLabel _ mclocation _ count) ->
+          case mclocation of
+            Nothing ->
+              pure Nothing
+            Just clocation ->
+              let
+                coverageMsg =
+                  WL.cat [
+                      "Failed ("
+                    , WL.annotate CoverageText $
+                        ppCoverPercentage (coverPercentage tests count) <> " coverage"
+                    , ")"
+                    ]
+              in
+                ppFailureLocation [coverageMsg] Nothing clocation
+
+  (args, idecls) <- fmap partitionEithers $ zipWithM ppFailedInput [0..] inputs0
 
   let
     decls =
       mergeDeclarations .
       catMaybes $
-        mlocation : fmap pure idecls
+        mlocation : coverageLocations <> fmap pure idecls
 
     with xs f =
       if null xs then
@@ -655,14 +663,31 @@ ppFailureReport name (FailureReport size seed _ inputs0 mlocation0 msg mdiff msg
       else
         [f xs]
 
-  pure . WL.indent 2 . WL.vsep . WL.punctuate WL.line $ concat [
+    whenSome f xs =
+      if null xs then
+        xs
+      else
+        f xs
+
+    bottom =
+      maybe [ppReproduce name size seed] (const []) mcoverage
+
+  pure .
+    whenSome (mempty :) .
+    whenSome (++ [mempty]) .
+    WL.punctuate WL.line .
+    fmap (WL.vsep . fmap (WL.indent 2)) .
+    fmap (id :: [Doc Markup] -> [Doc Markup]) .
+    filter (not . null) $
+    concat [
       with args $
-        WL.vsep . WL.punctuate WL.line
+        WL.punctuate WL.line
     , with decls $
-        WL.vsep . WL.punctuate WL.line . fmap ppDeclaration
-    , with msgs $
-        WL.vsep
-    , [ppReproduce name size seed]
+        WL.punctuate WL.line . fmap ppDeclaration
+    , with msgs1 $
+        id
+    , with bottom $
+        id
     ]
 
 ppName :: Maybe PropertyName -> Doc a
@@ -673,56 +698,316 @@ ppName = \case
     WL.text name
 
 ppProgress :: MonadIO m => Maybe PropertyName -> Report Progress -> m (Doc Markup)
-ppProgress name (Report tests discards status) =
+ppProgress name (Report tests discards coverage status) =
   case status of
     Running ->
-      pure . icon RunningIcon '●' . WL.annotate RunningHeader $
-        ppName name <+>
-        "passed" <+>
-        ppTestCount tests <>
-        ppWithDiscardCount discards <+>
-        "(running)"
+      pure . WL.vsep $ [
+          icon RunningIcon '●' . WL.annotate RunningHeader $
+            ppName name <+>
+            "passed" <+>
+            ppTestCount tests <>
+            ppWithDiscardCount discards <+>
+            "(running)"
+        ] ++
+        ppCoverage tests coverage
 
     Shrinking failure ->
       pure . icon ShrinkingIcon '↯' . WL.annotate ShrinkingHeader $
         ppName name <+>
-        "failed after" <+>
+        "failed" <+> ppFailedAtLocation (failureLocation failure) <#>
+        "after" <+>
         ppTestCount tests <>
         ppShrinkDiscard (failureShrinks failure) discards <+>
         "(shrinking)"
 
 ppResult :: MonadIO m => Maybe PropertyName -> Report Result -> m (Doc Markup)
-ppResult name (Report tests discards result) =
+ppResult name (Report tests discards coverage result) = do
   case result of
     Failed failure -> do
-      pfailure <- ppFailureReport name failure
+      pfailure <- ppFailureReport name tests failure
       pure . WL.vsep $ [
-          icon FailedIcon '✗' . WL.annotate FailedHeader $
+          icon FailedIcon '✗' . WL.align . WL.annotate FailedText $
             ppName name <+>
-            "failed after" <+>
+            "failed" <+> ppFailedAtLocation (failureLocation failure) <#>
+            "after" <+>
             ppTestCount tests <>
             ppShrinkDiscard (failureShrinks failure) discards <>
             "."
-        , mempty
-        , pfailure
-        , mempty
-        ]
+        ] ++
+        ppCoverage tests coverage ++
+        pfailure
 
     GaveUp ->
-      pure . icon GaveUpIcon '⚐' . WL.annotate GaveUpHeader $
-        ppName name <+>
-        "gave up after" <+>
-        ppDiscardCount discards <>
-        ", passed" <+>
-        ppTestCount tests <>
-        "."
+      pure . WL.vsep $ [
+          icon GaveUpIcon '⚐' . WL.annotate GaveUpText $
+            ppName name <+>
+            "gave up after" <+>
+            ppDiscardCount discards <>
+            ", passed" <+>
+            ppTestCount tests <>
+            "."
+        ] ++
+        ppCoverage tests coverage
 
     OK ->
-      pure . icon SuccessIcon '✓' . WL.annotate SuccessHeader $
-        ppName name <+>
-        "passed" <+>
-        ppTestCount tests <>
-        "."
+      pure . WL.vsep $ [
+          icon SuccessIcon '✓' . WL.annotate SuccessText $
+            ppName name <+>
+            "passed" <+>
+            ppTestCount tests <>
+            "."
+        ] ++
+        ppCoverage tests coverage
+
+ppFailedAtLocation :: Maybe Span -> Doc Markup
+ppFailedAtLocation = \case
+  Just x ->
+    "at" <+>
+    WL.text (spanFile x) <> ":" <>
+    WL.pretty (unLineNo (spanStartLine x)) <> ":" <>
+    WL.pretty (unColumnNo (spanStartColumn x))
+  Nothing ->
+    mempty
+
+ppCoverage :: TestCount -> Coverage CoverCount -> [Doc Markup]
+ppCoverage tests x =
+  if Map.null (coverageLabels x) then
+    mempty
+  else
+    fmap (ppLabel tests (coverageWidth tests x)) .
+    List.sortOn labelLocation $
+    Map.elems (coverageLabels x)
+
+data ColumnWidth =
+  ColumnWidth {
+      widthPercentage :: !Int
+    , widthMinimum :: !Int
+    , widthName :: !Int
+    , _widthNameFail :: !Int
+    }
+
+instance Semigroup ColumnWidth where
+  (<>) (ColumnWidth p0 m0 n0 f0) (ColumnWidth p1 m1 n1 f1) =
+    ColumnWidth
+      (max p0 p1)
+      (max m0 m1)
+      (max n0 n1)
+      (max f0 f1)
+
+instance Monoid ColumnWidth where
+  mempty =
+    ColumnWidth 0 0 0 0
+  mappend =
+    (<>)
+
+coverageWidth :: TestCount -> Coverage CoverCount -> ColumnWidth
+coverageWidth tests (Coverage labels) =
+  foldMap (labelWidth tests) labels
+
+labelWidth :: TestCount -> Label CoverCount -> ColumnWidth
+labelWidth tests x =
+  let
+    percentage =
+      length .
+      renderCoverPercentage .
+      coverPercentage tests $
+      labelAnnotation x
+
+    minimum_ =
+      if labelMinimum x == 0 then
+        0
+      else
+        length .
+        renderCoverPercentage $
+        labelMinimum x
+
+    name =
+      length .
+      unLabelName $
+      labelName x
+
+    nameFail =
+      if labelCovered tests x then
+        0
+      else
+        name
+  in
+    ColumnWidth percentage minimum_ name nameFail
+
+ppLeftPad :: Int -> Doc a -> Doc a
+ppLeftPad n doc =
+  let
+    ndoc =
+      length (show doc)
+
+    pad =
+      WL.text $
+        List.replicate (n - ndoc) ' '
+  in
+    pad <> doc
+
+ppLabel :: TestCount -> ColumnWidth -> Label CoverCount -> Doc Markup
+ppLabel tests w x@(MkLabel name _ minimum_ count) =
+  let
+    covered =
+      labelCovered tests x
+
+    ltext =
+      if not covered then
+        WL.annotate CoverageText
+      else
+        id
+
+    lborder =
+      WL.annotate (StyledBorder StyleDefault)
+
+    licon =
+      if not covered then
+        WL.annotate CoverageText "⚠ "
+      else
+        "  "
+
+    lname =
+      WL.fill (widthName w) (ppLabelName name)
+
+    wminimum =
+      ppLeftPad (widthMinimum w) $
+        ppCoverPercentage minimum_
+
+    wcover i =
+      ppLeftPad (widthPercentage w + length i) $
+        WL.text i <>
+        ppCoverPercentage (coverPercentage tests count)
+
+    lminimum =
+      if widthMinimum w == 0 then
+        mempty
+      else if not covered then
+        " ✗ " <> wminimum
+      else if minimum_ == 0 then
+        "   " <> ppLeftPad (widthMinimum w) ""
+      else
+        " ✓ " <> wminimum
+
+    lcover =
+      if widthMinimum w == 0 then
+        wcover ""
+      else if not covered then
+        wcover ""
+      else if minimum_ == 0 then
+        wcover ""
+      else
+        wcover ""
+  in
+    WL.hcat [
+        licon
+      , ltext lname
+      , lborder " "
+      , ltext lcover
+      , lborder " "
+      , ltext $ ppCoverBar (coverPercentage tests count) minimum_
+      , lborder "" -- "│"
+      , ltext lminimum
+      ]
+
+ppLabelName :: LabelName -> Doc a
+ppLabelName (LabelName name) =
+  WL.text name
+
+ppCoverPercentage :: CoverPercentage -> Doc Markup
+ppCoverPercentage =
+  WL.text . renderCoverPercentage
+
+ppCoverBar :: CoverPercentage -> CoverPercentage -> Doc Markup
+ppCoverBar (CoverPercentage percentage) (CoverPercentage minimum_) =
+  let
+    barWidth :: Int
+    barWidth =
+      20
+
+    coverageRatio :: Double
+    coverageRatio =
+      percentage / 100.0
+
+    coverageWidth_ :: Int
+    coverageWidth_ =
+      floor $
+        coverageRatio * fromIntegral barWidth
+
+    minimumRatio :: Double
+    minimumRatio =
+      minimum_ / 100.0
+
+    minimumWidth :: Int
+    minimumWidth =
+      floor $
+        minimumRatio * fromIntegral barWidth
+
+    index :: [a] -> Int
+    index xs =
+      floor $
+        ((coverageRatio * fromIntegral barWidth) - fromIntegral coverageWidth_) *
+        fromIntegral (length xs)
+
+    part xs =
+      xs !! index xs
+
+    fillWidth =
+      barWidth - coverageWidth_ - 1
+
+    fillErrorWidth =
+      max 0 (minimumWidth - coverageWidth_ - 1)
+
+    fillSurplusWidth =
+      fillWidth - fillErrorWidth
+
+    bar :: (Char, [Char]) -> Doc Markup
+    bar (full, parts) =
+      WL.hcat [
+        WL.text $ replicate coverageWidth_ full
+      , if fillWidth >= 0 then
+          if index parts == 0 then
+            if fillErrorWidth > 0 then
+              WL.annotate FailedText $ WL.text [part parts]
+            else
+              WL.annotate CoverageFill $ WL.text [part parts]
+          else
+            WL.text [part parts]
+        else
+          ""
+      , WL.annotate FailedText . WL.text $
+          replicate fillErrorWidth (head parts)
+      , WL.annotate CoverageFill . WL.text $
+          replicate fillSurplusWidth (head parts)
+      --
+      -- Uncomment when debugging:
+      --
+      -- , WL.annotate CoverageFill . WL.text $
+      --        " " ++ show barWidth
+      --     ++ " " ++ show coverageWidth_
+      --     ++ " " ++ show minimumWidth
+      --     ++ " " ++ "/"
+      --     ++ " " ++ show fillErrorWidth
+      --     ++ " " ++ "+"
+      --     ++ " " ++ show fillSurplusWidth
+      --     ++ " " ++ "="
+      --     ++ " " ++ show fillWidth
+      ]
+  in
+    bar ('█', ['·', '▏', '▎', '▍', '▌', '▋', '▊', '▉'])
+
+    -- FIXME Maybe this should be configurable?
+    -- Alternative histogram bars:
+    --bar ('⣿', ['·', '⡀', '⡄', '⡆', '⡇', '⣇', '⣧', '⣷'])
+    --bar ('⣿', ['⢕', '⡀', '⣀', '⣄', '⣤', '⣦', '⣶', '⣷'])
+    --bar ('⣿', ['⢕', '⡵', '⢗', '⣗', '⣟'])
+    --bar ('⣿', [' ', '⡵', '⢗', '⣗', '⣟'])
+    --bar ('█', ['░','▓'])
+    --bar ('█', ['░'])
+
+renderCoverPercentage :: CoverPercentage -> String
+renderCoverPercentage (CoverPercentage percentage) =
+  printf "%.0f" percentage <> "%"
 
 ppWhenNonZero :: Doc a -> PropertyCount -> Maybe (Doc a)
 ppWhenNonZero suffix n =
@@ -734,13 +1019,13 @@ ppWhenNonZero suffix n =
 annotateSummary :: Summary -> Doc Markup -> Doc Markup
 annotateSummary summary =
   if summaryFailed summary > 0 then
-    icon FailedIcon '✗' . WL.annotate FailedHeader
+    icon FailedIcon '✗' . WL.annotate FailedText
   else if summaryGaveUp summary > 0 then
-    icon GaveUpIcon '⚐' . WL.annotate GaveUpHeader
+    icon GaveUpIcon '⚐' . WL.annotate GaveUpText
   else if summaryWaiting summary > 0 || summaryRunning summary > 0 then
     icon WaitingIcon '○' . WL.annotate WaitingHeader
   else
-    icon SuccessIcon '✓' . WL.annotate SuccessHeader
+    icon SuccessIcon '✓' . WL.annotate SuccessText
 
 ppSummary :: MonadIO m => Summary -> m (Doc Markup)
 ppSummary summary =
@@ -784,8 +1069,8 @@ ppSummary summary =
             Nothing
         ]
 
-renderDoc :: MonadIO m => Maybe UseColor -> Doc Markup -> m String
-renderDoc mcolor doc = do
+renderDoc :: MonadIO m => UseColor -> Doc Markup -> m String
+renderDoc color doc = do
   let
     dull =
       SetColor Foreground Dull
@@ -811,16 +1096,22 @@ renderDoc mcolor doc = do
         setSGRCode [vivid Red]
       FailedIcon ->
         setSGRCode [vivid Red]
-      FailedHeader ->
+      FailedText ->
         setSGRCode [vivid Red]
       GaveUpIcon ->
         setSGRCode [dull Yellow]
-      GaveUpHeader ->
+      GaveUpText ->
         setSGRCode [dull Yellow]
       SuccessIcon ->
         setSGRCode [dull Green]
-      SuccessHeader ->
+      SuccessText ->
         setSGRCode [dull Green]
+      CoverageIcon ->
+        setSGRCode [dull Yellow]
+      CoverageText ->
+        setSGRCode [dull Yellow]
+      CoverageFill ->
+        setSGRCode [vivid Black]
 
       DeclarationLocation ->
         setSGRCode []
@@ -879,8 +1170,6 @@ renderDoc mcolor doc = do
     end _ =
       setSGRCode [Reset]
 
-  color <- resolveColor mcolor
-
   let
     display =
       case color of
@@ -894,19 +1183,20 @@ renderDoc mcolor doc = do
     hSetEncoding stdout utf8
     hSetEncoding stderr utf8
 #endif
+
   pure .
     display .
     WL.renderSmart 100 $
     WL.indent 2 doc
 
-renderProgress :: MonadIO m => Maybe UseColor -> Maybe PropertyName -> Report Progress -> m String
-renderProgress mcolor name x =
-  renderDoc mcolor =<< ppProgress name x
+renderProgress :: MonadIO m => UseColor -> Maybe PropertyName -> Report Progress -> m String
+renderProgress color name x =
+  renderDoc color =<< ppProgress name x
 
-renderResult :: MonadIO m => Maybe UseColor -> Maybe PropertyName -> Report Result -> m String
-renderResult mcolor name x =
-  renderDoc mcolor =<< ppResult name x
+renderResult :: MonadIO m => UseColor -> Maybe PropertyName -> Report Result -> m String
+renderResult color name x =
+  renderDoc color =<< ppResult name x
 
-renderSummary :: MonadIO m => Maybe UseColor -> Summary -> m String
-renderSummary mcolor x =
-  renderDoc mcolor =<< ppSummary x
+renderSummary :: MonadIO m => UseColor -> Summary -> m String
+renderSummary color x =
+  renderDoc color =<< ppSummary x
