@@ -56,27 +56,27 @@ import qualified Control.Concurrent.Async.Lifted as Async
 import           Control.Monad (foldM, foldM_)
 import           Control.Monad.Catch (MonadCatch)
 import           Control.Monad.State.Class (MonadState, get, put, modify)
+import           Control.Monad.Morph (MFunctor(..))
 import           Control.Monad.Trans.Class (lift)
 import           Control.Monad.Trans.Control (MonadBaseControl)
 import           Control.Monad.Trans.State (State, runState, execState)
-import           Control.Monad.Trans.State (StateT(..), evalStateT)
+import           Control.Monad.Trans.State (StateT(..), evalStateT, runStateT)
 
 import           Data.Dynamic (Dynamic, toDyn, fromDynamic, dynTypeRep)
 import           Data.Foldable (traverse_)
 import           Data.Functor.Classes (Eq1(..), Ord1(..), Show1(..))
-#if MIN_VERSION_transformers(0,5,0)
 import           Data.Functor.Classes (eq1, compare1, showsPrec1)
-#endif
 import           Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Maybe as Maybe
 import           Data.Typeable (Typeable, TypeRep, Proxy(..), typeRep)
 
-import           Hedgehog.Internal.Gen (MonadGen)
+import           Hedgehog.Internal.Distributive (distributeT)
+import           Hedgehog.Internal.Gen (MonadGen, GenT, GenBase)
 import qualified Hedgehog.Internal.Gen as Gen
 import           Hedgehog.Internal.HTraversable (HTraversable(..))
 import           Hedgehog.Internal.Opaque (Opaque(..))
-import           Hedgehog.Internal.Property (MonadTest(..), Test, evalEither, evalM, success, runTest, failWith)
+import           Hedgehog.Internal.Property (MonadTest(..), Test, evalEither, evalM, success, runTest, failWith, annotate)
 import           Hedgehog.Internal.Range (Range)
 import           Hedgehog.Internal.Show (showPretty)
 import           Hedgehog.Internal.Source (HasCallStack, withFrozenCallStack)
@@ -92,7 +92,17 @@ instance Show Name where
   showsPrec p (Name x) =
     showsPrec p x
 
--- | Symbolic values.
+-- | Symbolic values: Because hedgehog generates actions in a separate phase
+--   before execution, you will sometimes need to refer to the result of a
+--   previous action in a generator without knowing the value of the result
+--   (e.g., to get the ID of a previously-created user).
+--
+--   Symbolic variables provide a token to stand in for the actual variables at
+--   generation time (and in 'Require'/'Update' callbacks). At execution time,
+--   real values are available, so your execute actions work on 'Concrete'
+--   variables.
+--
+--   See also: 'Command', 'Var'
 --
 data Symbolic a where
   Symbolic :: Typeable a => Name -> Symbolic a
@@ -104,7 +114,6 @@ instance Show (Symbolic a) where
   showsPrec p (Symbolic x) =
     showsPrec p x
 
-#if MIN_VERSION_transformers(0,5,0)
 instance Show1 Symbolic where
   liftShowsPrec _ _ p (Symbolic x) =
     showsPrec p x
@@ -116,21 +125,11 @@ instance Eq1 Symbolic where
 instance Ord1 Symbolic where
   liftCompare _ (Symbolic x) (Symbolic y) =
     compare x y
-#else
-instance Show1 Symbolic where
-  showsPrec1 p (Symbolic x) =
-    showsPrec p x
 
-instance Eq1 Symbolic where
-  eq1 (Symbolic x) (Symbolic y) =
-    x == y
-
-instance Ord1 Symbolic where
-  compare1 (Symbolic x) (Symbolic y) =
-    compare x y
-#endif
-
--- | Concrete values.
+-- | Concrete values: At test-execution time, 'Symbolic' values from generation
+--   are replaced with 'Concrete' values from performing actions. This type
+--   gives us something of the same kind as 'Symbolic' to pass as a type
+--   argument to 'Var'.
 --
 newtype Concrete a where
   Concrete :: a -> Concrete a
@@ -140,7 +139,6 @@ instance Show a => Show (Concrete a) where
   showsPrec =
     showsPrec1
 
-#if MIN_VERSION_transformers(0,5,0)
 instance Show1 Concrete where
   liftShowsPrec sp _ p (Concrete x) =
     sp p x
@@ -152,19 +150,6 @@ instance Eq1 Concrete where
 instance Ord1 Concrete where
   liftCompare comp (Concrete x) (Concrete y) =
     comp x y
-#else
-instance Show1 Concrete where
-  showsPrec1 p (Concrete x) =
-    showsPrec p x
-
-instance Eq1 Concrete where
-  eq1 (Concrete x) (Concrete y) =
-    x == y
-
-instance Ord1 Concrete where
-  compare1 (Concrete x) (Concrete y) =
-    compare x y
-#endif
 
 ------------------------------------------------------------------------
 
@@ -185,7 +170,10 @@ instance Ord1 Concrete where
 --   type of variable because it is used in both the generation and the
 --   execution phase.
 --
-data Var a v =
+--   The order of arguments makes 'Var' 'HTraverable', which is how 'Symbolic'
+--   values are turned into 'Concrete' ones.
+--
+newtype Var a v =
   Var (v a)
 
 -- | Take the value from a concrete variable.
@@ -241,6 +229,10 @@ emptyEnvironment :: Environment
 emptyEnvironment =
   Environment Map.empty
 
+unionsEnvironment :: [Environment] -> Environment
+unionsEnvironment =
+  Environment . Map.unions . fmap unEnvironment
+
 -- | Insert a symbolic / concrete pairing in to the environment.
 --
 insertConcrete :: Symbolic a -> Concrete a -> Environment -> Environment
@@ -375,18 +367,24 @@ callbackEnsure callbacks s0 s i o =
 
 ------------------------------------------------------------------------
 
--- | The specification for the expected behaviour of an 'Action'.
+-- | The specification for the expected behaviour of an
+-- 'Action'. These are used to generate sequences of actions to test.
 --
-data Command n m (state :: (* -> *) -> *) =
+-- This is the main type you will use when writing state machine
+-- tests. @gen@ is usually an instance of 'MonadGen', and @m@ is usually
+-- an instance of 'MonadTest'. These constraints appear when you pass
+-- your 'Command' list to 'sequential' or 'parallel'.
+--
+data Command gen m (state :: (* -> *) -> *) =
   forall input output.
-  (HTraversable input, Show (input Symbolic), Typeable output) =>
+  (HTraversable input, Show (input Symbolic), Show output, Typeable output) =>
   Command {
     -- | A generator which provides random arguments for a command. If the
     --   command cannot be executed in the current state, it should return
     --   'Nothing'.
     --
       commandGen ::
-        state Symbolic -> Maybe (n (input Symbolic))
+        state Symbolic -> Maybe (gen (input Symbolic))
 
     -- | Executes a command using the arguments generated by 'commandGen'.
     --
@@ -402,7 +400,7 @@ data Command n m (state :: (* -> *) -> *) =
 
 -- | Checks that input for a command can be executed in the given state.
 --
-commandGenOK :: Command n m state -> state Symbolic -> Bool
+commandGenOK :: Command gen m state -> state Symbolic -> Bool
 commandGenOK (Command inputGen _ _) state =
   Maybe.isJust (inputGen state)
 
@@ -411,7 +409,7 @@ commandGenOK (Command inputGen _ _) state =
 --
 data Action m (state :: (* -> *) -> *) =
   forall input output.
-  (HTraversable input, Show (input Symbolic)) =>
+  (HTraversable input, Show (input Symbolic), Show output) =>
   Action {
       actionInput ::
         input Symbolic
@@ -536,11 +534,11 @@ dropInvalid =
 -- | Generates a single action from a set of possible commands.
 --
 action ::
-     (MonadGen n, MonadTest m)
-  => [Command n m state]
-  -> StateT (Context state) n (Action m state)
+     (MonadGen gen, MonadTest m)
+  => [Command gen m state]
+  -> GenT (StateT (Context state) (GenBase gen)) (Action m state)
 action commands =
-  Gen.just $ do
+  Gen.justT $ do
     Context state0 _ <- get
 
     Command mgenInput exec callbacks <-
@@ -550,8 +548,8 @@ action commands =
       case mgenInput state0 of
         Nothing ->
           error "genCommand: internal error, tried to use generator with invalid state."
-        Just g ->
-          lift g
+        Just gen ->
+          hoist lift $ Gen.toGenT gen
 
     if not $ callbackRequire callbacks state0 input then
       pure Nothing
@@ -569,13 +567,13 @@ action commands =
           (callbackEnsure callbacks)
 
 genActions ::
-     (MonadGen n, MonadTest m)
+     (MonadGen gen, MonadTest m)
   => Range Int
-  -> [Command n m state]
+  -> [Command gen m state]
   -> Context state
-  -> n ([Action m state], Context state)
+  -> gen ([Action m state], Context state)
 genActions range commands ctx = do
-  xs <- Gen.list range (action commands) `evalStateT` ctx
+  xs <- Gen.fromGenT . (`evalStateT` ctx) . distributeT $ Gen.list range (action commands)
   pure $
     dropInvalid xs `runState` ctx
 
@@ -603,6 +601,33 @@ renderAction (Action input (Symbolic (Name output)) _ _ _ _) =
         (prefix0 ++ x) :
         fmap (prefix ++) xs
 
+renderActionResult :: Environment -> Action m state -> [String]
+renderActionResult env (Action _ output@(Symbolic (Name name)) _ _ _ _) =
+  let
+    prefix0 =
+      "Var " ++ show name ++ " = "
+
+    prefix =
+      replicate (length prefix0) ' '
+
+    unfound = \case
+      EnvironmentValueNotFound _
+        -> "<<not found in environment>>"
+      EnvironmentTypeError _ _
+        -> "<<type representation in environment unexpected>>"
+
+    actual =
+      either unfound showPretty
+        $ reifyEnvironment env output
+
+  in
+    case lines actual of
+      [] ->
+        [prefix0 ++ "?"]
+      x : xs ->
+        (prefix0 ++ x) :
+        fmap (prefix ++) xs
+
 -- FIXME we should not abuse Show to get nice output for actions
 instance Show (Sequential m state) where
   show (Sequential xs) =
@@ -611,11 +636,11 @@ instance Show (Sequential m state) where
 -- | Generates a sequence of actions from an initial model state and set of commands.
 --
 sequential ::
-     (MonadGen n, MonadTest m)
+     (MonadGen gen, MonadTest m)
   => Range Int
   -> (forall v. state v)
-  -> [Command n m state]
-  -> n (Sequential m state)
+  -> [Command gen m state]
+  -> gen (Sequential m state)
 sequential range initial commands =
   fmap (Sequential . fst) $
     genActions range commands (mkContext initial)
@@ -636,27 +661,32 @@ data Parallel m state =
 
 -- FIXME we should not abuse Show to get nice output for actions
 instance Show (Parallel m state) where
-  show (Parallel pre xs ys) =
-    unlines $ concat [
-        ["━━━ Prefix ━━━"]
-      ,  (concatMap renderAction pre)
-      , ["", "━━━ Branch 1 ━━━"]
-      ,  (concatMap renderAction xs)
-      , ["", "━━━ Branch 2 ━━━"]
-      ,  (concatMap renderAction ys)
-      ]
+  show =
+    renderParallel renderAction
+
+renderParallel :: (Action m state -> [String]) -> Parallel m state -> String
+renderParallel render (Parallel pre xs ys) =
+  unlines $ concat [
+      ["━━━ Prefix ━━━"]
+    , concatMap render pre
+    , ["", "━━━ Branch 1 ━━━"]
+    , concatMap render xs
+    , ["", "━━━ Branch 2 ━━━"]
+    , concatMap render ys
+    ]
+
 
 -- | Given the initial model state and set of commands, generates prefix
 --   actions to be run sequentially, followed by two branches to be run in
 --   parallel.
 --
 parallel ::
-     (MonadGen n, MonadTest m)
+     (MonadGen gen, MonadTest m)
   => Range Int
   -> Range Int
   -> (forall v. state v)
-  -> [Command n m state]
-  -> n (Parallel m state)
+  -> [Command gen m state]
+  -> gen (Parallel m state)
 parallel prefixN parallelN initial commands = do
   (prefix, ctx0) <- genActions prefixN commands (mkContext initial)
   (branch1, ctx1) <- genActions parallelN commands ctx0
@@ -792,13 +822,17 @@ executeParallel ::
   => (forall v. state v)
   -> Parallel m state
   -> m ()
-executeParallel initial (Parallel prefix branch1 branch2) =
+executeParallel initial p@(Parallel prefix branch1 branch2) =
   withFrozenCallStack $ evalM $ do
     (s0, env0) <- foldM executeUpdateEnsure (initial, emptyEnvironment) prefix
 
-    (xs, ys) <-
+    ((xs, env1), (ys, env2)) <-
       Async.concurrently
-        (evalStateT (traverse execute branch1) env0)
-        (evalStateT (traverse execute branch2) env0)
+        (runStateT (traverse execute branch1) env0)
+        (runStateT (traverse execute branch2) env0)
 
+    let
+      env = unionsEnvironment [env0, env1, env2]
+
+    annotate $ renderParallel (renderActionResult env) p
     linearize s0 xs ys
